package guavacado;

import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javafx.animation.Interpolator;
import javafx.animation.KeyFrame;
import javafx.animation.KeyValue;
import javafx.animation.Timeline;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.geometry.Insets;
import javafx.geometry.Pos;
import javafx.scene.Scene;
import javafx.scene.control.Button;
import javafx.scene.control.ChoiceBox;
import javafx.scene.control.Label;
import javafx.scene.control.ScrollPane;
import javafx.scene.control.Slider;
import javafx.scene.control.Tooltip;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.layout.HBox;
import javafx.scene.layout.Pane;
import javafx.scene.layout.Priority;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.scene.shape.Arc;
import javafx.scene.shape.ArcType;
import javafx.scene.shape.Circle;
import javafx.scene.shape.StrokeType;
import javafx.stage.DirectoryChooser;
import javafx.stage.FileChooser;
import javafx.stage.Modality;
import javafx.stage.Stage;
import javafx.stage.StageStyle;
import javafx.stage.WindowEvent;
import javafx.util.Duration;

/**
 * Guavacado is an application for batch resizing/exporting images to the
 * ico, png, jpg or bmp format.
 * 
 * Feel free to distribute, modify and use this code in your own projects
 * without restriction.
 * 
 * @author Adam T. Ryder http://1337atr.weebly.com
 * @version 1.0
 *
 */
public class GuavacadoMainWin {
	public static final float VERSION = 1.01f;
	
	private static final int popupAnim = 154;
	
	public static final int PROGRESSTYPE_SAVEPROJECT = 0;
	public static final int PROGRESSTYPE_OPENPROJECT = 1;
	public static final int PROGRESSTYPE_EXPORTBATCH = 2;
	public static final int PROGRESSTYPE_EXPORTICO = 3;
	
	public static final Image avacadoToggleLockedSmall = new Image("/guavacado/resources/gfx/Button_ToggleLocked_Small.png");
	public static final Image avacadoToggleUnlockedSmall = new Image("/guavacado/resources/gfx/Button_ToggleUnlocked_Small.png");
	
	private Stage primaryStage;
	
	@FXML
	private Pane buttonPane;
	@FXML
	private ScrollPane iconScrollArea;
	@FXML
	private Pane listPane;
	
	private final Timeline timeLine = new Timeline();
	private final Timeline progressTimeline = new Timeline();
	
	private final List<IconTile> tiles = new ArrayList<IconTile>();
	
	private File currentProject = null;
	
	private boolean loading = false;
	
	private int lastFormatSelection = 0;
	private Double lastQualitySelection = 90.0;
	
	private File lastSave = new File(System.getProperty("user.home"));
	private File lastOpen = new File(System.getProperty("user.home"));
	private File lastImport = new File(System.getProperty("user.home"));
	private File lastExportICO = new File(System.getProperty("user.home"));
	private File lastExport = new File(System.getProperty("user.home"));
	
	private final File prefFile = new File(System.getProperty("user.home"), "GuavacadoSettings.guav");
	
	private List<Image> icons = new ArrayList<Image>();
	
	/**
	 * Called from Main during initial application setup to pass the primary stage to this
	 * main window controller, read the preferences file, setup intro animation, setup handlers
	 * for close requests and read any files passed as command line arguments.
	 * 
	 * @param primaryStage The primary stage for Guavacado.
	 * @param args Paths to supported files passed as command line arguments.
	 */
	protected void prepare(final Stage primaryStage, final List<String> args) {
		ImageLoader.mainWin = this;
		this.primaryStage = primaryStage;
		
		iconScrollArea.setHmax(0);
		iconScrollArea.setHmin(0);
		
		icons.add(new Image("/guavacado/resources/gfx/GuavacadoIcon_72_16_32bit.png"));
		icons.add(new Image("/guavacado/resources/gfx/GuavacadoIcon_72_24_32bit.png"));
		icons.add(new Image("/guavacado/resources/gfx/GuavacadoIcon_72_32_32bit.png"));
		icons.add(new Image("/guavacado/resources/gfx/GuavacadoIcon_72_48_32bit.png"));
		icons.add(new Image("/guavacado/resources/gfx/GuavacadoIcon_72_64_32bit.png"));
		icons.add(new Image("/guavacado/resources/gfx/GuavacadoIcon_72_72_32bit.png"));
		
		progressTimeline.setCycleCount(1);
		
		if (prefFile.exists()) {
			DataInputStream dis = null;
			try {
				dis = new DataInputStream(new BufferedInputStream(new FileInputStream(prefFile)));
				float ver = dis.readFloat();
				String lSave = dis.readUTF();
				String lOpen = dis.readUTF();
				String lImport = dis.readUTF();
				String lExpICO = dis.readUTF();
				String lExp = dis.readUTF();
				int lastFormat = dis.readInt();
				double lastQuality = dis.readDouble();
				
				File file = new File(lSave);
				if (file.exists()) { lastSave = file; }
				file = new File(lOpen);
				if (file.exists()) { lastOpen = file; }
				file = new File(lImport);
				if (file.exists()) { lastImport = file; }
				file = new File(lExpICO);
				if (file.exists()) { lastExportICO = file; }
				file = new File(lExp);
				if (file.exists()) { lastExport = file; }
				lastFormatSelection = (lastFormat >= 0 && lastFormat <= 2) ? lastFormat : lastFormatSelection;
				lastQualitySelection = (lastQuality >= 0 && lastQuality <= 100) ? lastQuality : lastQualitySelection;
				
			} catch (Exception e) {
				System.out.println("Guavacado was unable to save the settings file:\n" + e.getMessage());
			} finally {
				if (dis != null) {
					try {
						dis.close();
					} catch (IOException ioe) {
						
					}
				}
			}
		}
		
		primaryStage.setOnCloseRequest(new EventHandler<WindowEvent>() {
			@Override
			public void handle(WindowEvent event) {
				ImageLoader.setCancelled();
				ImageExport.setCancelled();
				ProjectSave.setCancelled();
				ProjectOpen.setCancelled();
				for (IconTile tile : tiles) {
					tile.stopOperations();
				}
				timeLine.stop();
				progressTimeline.stop();
				
				DataOutputStream dos = null;
				try {
					dos = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(prefFile)));
					dos.writeFloat(VERSION);
					dos.writeUTF(lastSave.getPath());
					dos.writeUTF(lastOpen.getPath());
					dos.writeUTF(lastImport.getPath());
					dos.writeUTF(lastExportICO.getPath());
					dos.writeUTF(lastExport.getPath());
					dos.writeInt(lastFormatSelection);
					dos.writeDouble(lastQualitySelection);
				} catch (IOException e) {
					System.out.println("Guavacado was unable to save the settings file:\n" + e.getMessage());
				} finally {
					if (dos != null) {
						try {
							dos.close();
						} catch (IOException ioe) {
							
						}
					}
				}
			}
		});
		
		primaryStage.setOnShown(new EventHandler<WindowEvent>() {
			@Override
			public void handle(WindowEvent event) {
				timeLine.getKeyFrames().add(new KeyFrame(Duration.millis(600), new KeyValue(buttonPane.translateXProperty(), -1, Interpolator.EASE_OUT)));
				
				buttonPane.setDisable(true);
				
				((Button)buttonPane.getChildren().get(0)).setGraphic(new ImageView(new Image("/guavacado/resources/gfx/Button_NewProject.png")));
				((Button)buttonPane.getChildren().get(1)).setGraphic(new ImageView(new Image("/guavacado/resources/gfx/Button_OpenProject.png")));
				((Button)buttonPane.getChildren().get(2)).setGraphic(new ImageView(new Image("/guavacado/resources/gfx/Button_SaveProject.png")));
				((Button)buttonPane.getChildren().get(3)).setGraphic(new ImageView(new Image("/guavacado/resources/gfx/Button_SaveProjectAs.png")));
				((Button)buttonPane.getChildren().get(4)).setGraphic(new ImageView(new Image("/guavacado/resources/gfx/Button_Import.png")));
				((Button)buttonPane.getChildren().get(5)).setGraphic(new ImageView(new Image("/guavacado/resources/gfx/Button_ExportIco.png")));
				((Button)buttonPane.getChildren().get(6)).setGraphic(new ImageView(new Image("/guavacado/resources/gfx/Button_ExportBatch.png")));
				((Button)buttonPane.getChildren().get(7)).setGraphic(new ImageView(new Image("/guavacado/resources/gfx/Button_Help.png")));
				
				Tooltip.install(buttonPane.getChildren().get(0), new Tooltip("New Guavacado project"));
				Tooltip.install(buttonPane.getChildren().get(1), new Tooltip("Open Guavacado project"));
				Tooltip.install(buttonPane.getChildren().get(2), new Tooltip("Save Guavacado project"));
				Tooltip.install(buttonPane.getChildren().get(3), new Tooltip("Save As Guavacado project"));
				Tooltip.install(buttonPane.getChildren().get(4), new Tooltip("Import images"));
				Tooltip.install(buttonPane.getChildren().get(5), new Tooltip("Export project as an icon (.ico)"));
				Tooltip.install(buttonPane.getChildren().get(6), new Tooltip("Export project as separate images"));
				Tooltip.install(buttonPane.getChildren().get(7), new Tooltip("Get help with Guavacado"));
				int offset = (int)buttonPane.getChildren().get(buttonPane.getChildren().size() - 1).getTranslateY() + 64;
				for (int i = 0; i < buttonPane.getChildren().size(); i++) {
					Button b = (Button)buttonPane.getChildren().get(i);
					b.setFocusTraversable(false);
					timeLine.getKeyFrames().addAll(
							new KeyFrame(Duration.millis(400 + (83 * ((buttonPane.getChildren().size() - 1) - i))), new KeyValue(b.translateYProperty(), b.getTranslateY() - offset)),
							new KeyFrame(Duration.millis(1100 + (83 * ((buttonPane.getChildren().size() - 1) - i))), new KeyValue(b.translateYProperty(), b.getTranslateY(), Interpolator.EASE_OUT))
					);
					b.setTranslateY(b.getTranslateY() - offset);
				}
				timeLine.getKeyFrames().add(new KeyFrame(Duration.millis(1100 + (83 * (buttonPane.getChildren().size() - 1))), new EventHandler<ActionEvent>() {
					@Override
					public void handle(ActionEvent event) {
						buttonPane.setDisable(false);
						
						if (!args.isEmpty()) {
							List<File> files = new ArrayList<File>();
							for (String arg : args) {
								if (Main.isGPF(arg)) {
									files.clear();
									createProgressWindow(PROGRESSTYPE_OPENPROJECT, new File(arg));
									break;
								}
								files.add(new File(arg));
							}
							if (!files.isEmpty()) {
								beginImport(files);
							}
						}
					}
				}));
				buttonPane.setTranslateX(-buttonPane.getWidth());
				
				timeLine.setCycleCount(1);
				timeLine.play();
			}
		});
	}
	
	/**
	 * 
	 * @return File that represents the project currently being edited.
	 */
	protected File getCurrentProjectFile() {
		return currentProject;
	}
	
	/**
	 * 
	 * @return The <code>Pane</code> that contains all <code>IconTile</code>s.
	 */
	protected Pane getListPane() {
		return listPane;
	}
	
	/**
	 * 
	 * @return <code>List</code> of currently loaded <code>IconTile</code>s.
	 */
	protected List<IconTile> getTiles() {
		return tiles;
	}
	
	/**
	 * Removes an <code>IconTile</code> from the project.
	 * 
	 * @param tile The <code>IconTile</code> to be removed.
	 */
	protected void removeItem(final IconTile tile) {
		int index = tiles.indexOf(tile);
		tile.removeFromScene();
		tiles.remove(index);
		if (index < tiles.size()) {
			for (int i = index; i < tiles.size(); i++) {
				tiles.get(i).setNewYOffset(IconTile.height + IconTile.spacing);
			}
		} else {
			listPane.setMaxHeight((tiles.size() * (IconTile.height + IconTile.spacing)) + 48);
			listPane.setMinHeight((tiles.size() * (IconTile.height + IconTile.spacing)) + 48);
			listPane.setPrefHeight((tiles.size() * (IconTile.height + IconTile.spacing)) + 48);
		}
		buttonPane.requestFocus();
	}
	
	/**
	 * Creates a duplicate <code>IconTile</code> in the project.  Image, thumbnail & settings
	 * are copied to the new <code>IconTile</code>.
	 * 
	 * @param tile The <code>IconTile</code> to be duplicated.
	 */
	protected void cloneItem(final IconTile tile) {
		int index = tiles.indexOf(tile);
		if (index < tiles.size() - 1) {
			for (int i = index + 1; i < tiles.size(); i++) {
				tiles.get(i).setNewYOffset(-(IconTile.height + IconTile.spacing));
			}
			tiles.add(index + 1, new IconTile(tile, index, listPane, this));
		} else {
			tiles.add(new IconTile(tile, index, listPane, this));
		}
	}
	
	/**
	 * Clears the current project and starts a new project.
	 * 
	 * @param event The <code>ActionEvent</code> associated with the event that called
	 * this method, usually a button click.
	 */
	@FXML protected void newProject(ActionEvent event) {
		if (!tiles.isEmpty()) {
			currentProject = null;
			primaryStage.setTitle("Guavacado");
			clearProjectArea(null);
		}
	}
	
	/**
	 * Clears the <code>IconTile</code> area of the window and begins loading a series of image
	 * files if necessary.
	 * 
	 * @param filesToOpen Image files to load after clearing the current project.  Only clears the
	 * project without loading new files if null.
	 */
	private void clearProjectArea(ProjectOpen.projectImage[] filesToOpen) {
		if (loading) {
			for (IconTile tile : tiles) {
				tile.stopOperations();
			}
			loading = false;
		}
		buttonPane.setDisable(true);
		timeLine.stop();
		timeLine.getKeyFrames().clear();
		timeLine.getKeyFrames().addAll(
				new KeyFrame(Duration.ZERO, new KeyValue(listPane.opacityProperty(), 1)),
				new KeyFrame(Duration.millis(328), new KeyValue(listPane.opacityProperty(), 0, Interpolator.EASE_IN)),
				new KeyFrame(Duration.millis(339), new EventHandler<ActionEvent>() {
					@Override
					public void handle(ActionEvent event) {
						for (IconTile tile : tiles) {
							tile.stopOperations();
							listPane.getChildren().remove(tile.container);
						}
						tiles.clear();
						listPane.setOpacity(1);
						buttonPane.setDisable(false);
						listPane.setMaxHeight((IconTile.height + IconTile.spacing) + 48);
						listPane.setMinHeight((IconTile.height + IconTile.spacing) + 48);
						listPane.setPrefHeight((IconTile.height + IconTile.spacing) + 48);
						System.gc();
						
						if (filesToOpen != null) {
							beginImport(filesToOpen);
						}
					}
				})
		);
		timeLine.play();
		
		if (filesToOpen == null) {
			iconScrollArea.vvalueProperty().addListener(new ChangeListener<Number>() {
				@Override
				public void changed(ObservableValue<? extends Number> observable, Number oldVal, Number newVal) {
					if (newVal.doubleValue() == 1.0) {
						iconScrollArea.vvalueProperty().removeListener(this);
						iconScrollArea.setVvalue(oldVal.doubleValue());
					} else {
						iconScrollArea.vvalueProperty().removeListener(this);
					}
				}
			});
		}
	}
	
	/**
	 * Prompt user to select a project file then open the saved project.
	 * 
	 * @param event The <code>ActionEvent</code> associated with the event that called
	 * this method, usually a button click.
	 */
	@FXML protected void openProject(ActionEvent event) {
		if (!loading) {
			FileChooser fc = new FileChooser();
			fc.setTitle("Open Guavacado Project");
			if (lastOpen.exists()) {
				fc.setInitialDirectory(lastOpen);
			}
			fc.getExtensionFilters().addAll(new FileChooser.ExtensionFilter("Guavacado Project File", "*.gpf"));
			final File file = fc.showOpenDialog(primaryStage);
			
			if (file != null) {
				lastOpen = file.getParentFile();
				createProgressWindow(PROGRESSTYPE_OPENPROJECT, file);
			}
		} else {
			popMessage("FYI", "Guavacado is currently loading images.");
		}
	}
	
	/**
	 * Prompt user to select a file, if the project has not already been saved, then save the project.
	 * 
	 * @param event The <code>ActionEvent</code> associated with the event that called
	 * this method, usually a button click.
	 */
	@FXML protected void saveProject(ActionEvent event) {
		if (!tiles.isEmpty() && !loading) {
			if (currentProject == null) {
				FileChooser fc = new FileChooser();
				fc.setTitle("Save Guavacado Project");
				if (lastSave.exists()) {
					fc.setInitialDirectory(lastSave);
				}
				fc.getExtensionFilters().addAll(new FileChooser.ExtensionFilter("Guavacado Project File", "*.gpf"));
				File file = fc.showSaveDialog(primaryStage);
				
				if (file != null) {
					lastSave = file.getParentFile();
					currentProject = file;
					primaryStage.setTitle("Guavacado - " + currentProject.getName().substring(0, currentProject.getName().length() - 4));
					createProgressWindow(PROGRESSTYPE_SAVEPROJECT, currentProject);
				}
			} else {
				createProgressWindow(PROGRESSTYPE_SAVEPROJECT, currentProject);
			}
		} else if (loading) {
			popMessage("FYI", "Guavacado cannot complete this action while images are still loading.");
		} else if (tiles.isEmpty()) {
			popMessage("FYI", "There is nothing within the project to save.");
		}
	}
	
	/**
	 * Prompt user to select a file then save the project.
	 * 
	 * @param event The <code>ActionEvent</code> associated with the event that called
	 * this method, usually a button click.
	 */
	@FXML protected void saveProjectAs(ActionEvent event) {
		if (!tiles.isEmpty() && !loading) {
			FileChooser fc = new FileChooser();
			fc.setTitle("Save Guavacado Project As");
			if (lastSave.exists()) {
				fc.setInitialDirectory(lastSave);
			}
			fc.getExtensionFilters().addAll(new FileChooser.ExtensionFilter("Guavacado Project File", "*.gpf"));
			File file = fc.showSaveDialog(primaryStage);
			
			if (file != null) {
				lastSave = file.getParentFile();
				currentProject = file;
				primaryStage.setTitle("Guavacado - " + currentProject.getName().substring(0, currentProject.getName().length() - 4));
				createProgressWindow(PROGRESSTYPE_SAVEPROJECT, currentProject);
			}
		} else if (loading) {
			popMessage("FYI", "Guavacado is currently loading images.");
		} else if (tiles.isEmpty()) {
			popMessage("FYI", "There is nothing within the project to save.");
		}
	}
	
	/**
	 * Prompt user to select an image file, png/jpg/bmp, and import that image to the
	 * project.
	 * 
	 * @param event The <code>ActionEvent</code> associated with the event that called
	 * this method, usually a button click.
	 */
	@FXML protected void importImage(ActionEvent event) {
		if (!loading) {
			FileChooser fc = new FileChooser();
			fc.setTitle("Import Images");
			if (lastImport.exists()) {
				fc.setInitialDirectory(lastImport);
			}
			fc.getExtensionFilters().addAll(
					new FileChooser.ExtensionFilter("All Supported Formats", "*.png", "*.jpg", "*.jpeg", "*.bmp"),
					new FileChooser.ExtensionFilter("Portable Network Graphics", "*.png"),
					new FileChooser.ExtensionFilter("Joint Photographic Experts Group", "*.jpg", "*.jpeg"),
					new FileChooser.ExtensionFilter("Bitmap", "*.bmp")
			);
			List<File> files = fc.showOpenMultipleDialog(primaryStage);
			
			if (files != null) {
				lastImport = files.get(0).getParentFile();
				beginImport(files);
			}
		} else {
			popMessage("FYI", "Guavacado is currently loading images.");
		}
	}
	
	/**
	 * Begin importing images to the project.
	 * 
	 * @param images <code>List</code> of <code>File</code>s to import.
	 */
	protected void beginImport(final List<File> images) {
		beginImport(images.toArray(new File[images.size()]));
	}
	
	/**
	 * Begin importing images to the project.
	 * 
	 * @param images <code>Array</code> of <code>File</code>s to import.
	 */
	protected void beginImport(final File[] files) {
		IconTile newTile = new IconTile(files[0], this);
		newTile.addToScene(tiles.size(), listPane);
		tiles.add(newTile);
		listPane.setMaxHeight((tiles.size() * (IconTile.height + IconTile.spacing)) + 48);
		listPane.setMinHeight((tiles.size() * (IconTile.height + IconTile.spacing)) + 48);
		listPane.setPrefHeight((tiles.size() * (IconTile.height + IconTile.spacing)) + 48);
		
		if (files.length > 1) {
			loading = true;
			timeLine.stop();
			timeLine.getKeyFrames().clear();
			for (int i = 1; i < files.length; i++) {
				final File file = files[i];
				final GuavacadoMainWin thisClass = this;
				timeLine.getKeyFrames().add(new KeyFrame(Duration.millis(i * 142), new EventHandler<ActionEvent>() {
					@Override
					public void handle(ActionEvent event) {
						IconTile newTile = new IconTile(file, thisClass);
						newTile.addToScene(tiles.size(), listPane);
						tiles.add(newTile);
						listPane.setMaxHeight((tiles.size() * (IconTile.height + IconTile.spacing)) + 48);
						listPane.setMinHeight((tiles.size() * (IconTile.height + IconTile.spacing)) + 48);
						listPane.setPrefHeight((tiles.size() * (IconTile.height + IconTile.spacing)) + 48);
					}
				}));
			}
			timeLine.getKeyFrames().add(new KeyFrame(Duration.millis(files.length * 142), new EventHandler<ActionEvent>() {
				@Override
				public void handle(ActionEvent event) {
					loading = false;
				}
			}));
			timeLine.play();
		}
	}
	
	/**
	 * Begin importing images to the project.
	 * 
	 * @param projectImages <code>Array</code> of <code>ProjectOpen.projectImage</code>s to import.
	 * @see ProjectOpen.projectImage
	 */
	protected void beginImport(ProjectOpen.projectImage[] projectImages) {
		int index = -1;
		int numAdded = 0;
		for (int i = 0; i < projectImages.length; i++) {
			if (projectImages[i].isGood) {
				index = i;
				numAdded = 1;
				IconTile newTile = new IconTile(projectImages[0].path, this);
				newTile.addToScene(tiles.size(), listPane);
				newTile.setFromProjectLoad(projectImages[i]);
				tiles.add(newTile);
				listPane.setMaxHeight((tiles.size() * (IconTile.height + IconTile.spacing)) + 48);
				listPane.setMinHeight((tiles.size() * (IconTile.height + IconTile.spacing)) + 48);
				listPane.setPrefHeight((tiles.size() * (IconTile.height + IconTile.spacing)) + 48);
				break;
			}
		}
		
		if (projectImages.length > index + 1 && index >= 0) {
			loading = true;
			timeLine.stop();
			timeLine.getKeyFrames().clear();
			for (int i = index + 1; i < projectImages.length; i++) {
				final ProjectOpen.projectImage pi = projectImages[i];
				if (pi.isGood) {
					numAdded++;
					final GuavacadoMainWin thisClass = this;
					timeLine.getKeyFrames().add(new KeyFrame(Duration.millis(i * 142), new EventHandler<ActionEvent>() {
						@Override
						public void handle(ActionEvent event) {
							IconTile newTile = new IconTile(pi.path, thisClass);
							newTile.addToScene(tiles.size(), listPane);
							newTile.setFromProjectLoad(pi);
							tiles.add(newTile);
							listPane.setMaxHeight((tiles.size() * (IconTile.height + IconTile.spacing)) + 48);
							listPane.setMinHeight((tiles.size() * (IconTile.height + IconTile.spacing)) + 48);
							listPane.setPrefHeight((tiles.size() * (IconTile.height + IconTile.spacing)) + 48);
						}
					}));
				}
			}
			timeLine.getKeyFrames().add(new KeyFrame(Duration.millis(numAdded * 142), new EventHandler<ActionEvent>() {
				@Override
				public void handle(ActionEvent event) {
					loading = false;
				}
			}));
			timeLine.play();
		}
	}
	
	/**
	 * Prompt user to select a .ico file and begin scaling/exporting the project to the selected
	 * .ico file.
	 * 
	 * @param event The <code>ActionEvent</code> associated with the event that called
	 * this method, usually a button click.
	 */
	@FXML protected void exportIco(ActionEvent event) {
		if (!tiles.isEmpty() && !loading) {
			FileChooser fc = new FileChooser();
			fc.setTitle("Export Icon");
			if (lastExportICO.exists()) {
				fc.setInitialDirectory(lastExportICO);
			}
			fc.getExtensionFilters().add(new FileChooser.ExtensionFilter("Icon", "*.ico"));
			File file = fc.showSaveDialog(primaryStage);
			
			if (file != null) {
				lastExportICO = file.getParentFile();
				createProgressWindow(PROGRESSTYPE_EXPORTICO, file);
			}
		} else if (loading) {
			popMessage("FYI", "Guavacado is currently loading images.");
		} else if (tiles.isEmpty()) {
			popMessage("FYI", "There are no loaded images to export.");
		}
	}
	
	/**
	 * Prompt the user to select a directory and begin scaling/exporting the project images to
	 * the selected directory.
	 * 
	 * @param event The <code>ActionEvent</code> associated with the event that called
	 * this method, usually a button click.
	 */
	@FXML protected void export(ActionEvent event) {
		if (!tiles.isEmpty() && !loading) {
			DirectoryChooser dc = new DirectoryChooser();
			dc.setTitle("Export Images");
			if (lastExport.exists()) {
				dc.setInitialDirectory(lastExport);
			}
			File file = dc.showDialog(primaryStage);
			
			if (file != null) {
				lastExport = file;
				createExportDialog();
			}
		} else if (loading) {
			popMessage("FYI", "Guavacado is currently loading images.");
		} else if (tiles.isEmpty()) {
			popMessage("FYI", "There are no loaded images to export.");
		}
	}
	
	/**
	 * Open the help window.
	 * 
	 * @param event The <code>ActionEvent</code> associated with the event that called
	 * this method, usually a button click.
	 */
	@FXML protected void help(ActionEvent event) {
		try {
			final Stage dialogStage = new Stage(StageStyle.DECORATED);
			
			FXMLLoader loader = new FXMLLoader(Main.class.getResource("/guavacado/resources/fxml/GuavaHelp.fxml"));
			Pane dialogPane = (Pane) loader.load();
			((GuavHelp) loader.getController()).prepare(dialogStage, buttonPane.getChildren());
			
			Scene scene = new Scene(dialogPane);
			scene.getStylesheets().add(Main.class.getResource("/guavacado/resources/fxml/GuavacadoTheme.css").toExternalForm());
			dialogStage.setScene(scene);
			dialogStage.initOwner(primaryStage);
			dialogStage.initModality(Modality.APPLICATION_MODAL);
			dialogStage.getIcons().addAll(icons);
			dialogStage.setTitle("Guavacado Help");
			dialogStage.show();
		} catch (Exception e) {
			System.out.println("Unable to display help window due to the following error:");
			System.out.println(e.getMessage());
		}
	}
	
	/**
	 * Disables the main window, displays a progress indicator and begins a process associated with
	 * <code>progressType</code>.
	 * 
	 * @param progressType The type of process to begin once the progress window's intro animation
	 * is complete. Such as <code>PROGRESSTYPE_SAVEPROJECT</code>
	 * @param path The path to the file or directory being worked on.
	 */
	private void createProgressWindow(final int progressType, final File path) {
		buttonPane.setDisable(true);
		iconScrollArea.setDisable(true);
		
		double paneSize = 256;
		Pane rootContainer = new Pane();
		rootContainer.setPrefSize(paneSize, paneSize);
		rootContainer.setMaxSize(paneSize, paneSize);
		rootContainer.setMinSize(paneSize, paneSize);
		rootContainer.setTranslateX((primaryStage.getScene().getWidth() / 2) - (paneSize / 2));
		rootContainer.setTranslateY((primaryStage.getScene().getHeight() / 2) - (paneSize / 2));
		
		Pane expPane = new Pane();
		expPane.setLayoutX(0);
		expPane.setLayoutY(0);
		expPane.setPrefSize(paneSize, paneSize);
		expPane.setMaxSize(paneSize, paneSize);
		expPane.setMinSize(paneSize, paneSize);
		expPane.getStyleClass().add("popup-pane");
		rootContainer.getChildren().add(expPane);
		
		Circle arcBackground = new Circle(paneSize / 2f, paneSize / 2f, (paneSize / 2f) * 0.8f, Color.TRANSPARENT);
		arcBackground.setStroke(Color.DARKGREY);
		arcBackground.setStrokeWidth(15);
		arcBackground.setStrokeType(StrokeType.CENTERED);
		
		Arc arcProgress = new Arc(paneSize / 2f, paneSize / 2f, (paneSize / 2f) * 0.8f, (paneSize / 2f) * 0.8f, 90, 0);
		arcProgress.setType(ArcType.OPEN);
		arcProgress.setFill(Color.TRANSPARENT);
		arcProgress.setStroke(Color.CYAN);
		arcProgress.setStrokeWidth(11);
		arcProgress.setStrokeType(StrokeType.CENTERED);
		
		HBox textBox = new HBox();
		textBox.setPrefSize(paneSize, paneSize);
		textBox.setMinSize(paneSize, paneSize);
		textBox.setMaxSize(paneSize, paneSize);
		textBox.setAlignment(Pos.CENTER);
		
		Label progressLabel = new Label("0%");
		progressLabel.setStyle("-fx-text-fill: rgb(80, 80, 80);"
				+ "-fx-font-size: 22px;"
				+ "-fx-font-weight: bold;");
		textBox.getChildren().add(progressLabel);
		
		expPane.getChildren().addAll(arcBackground, arcProgress, textBox);
		((Pane)primaryStage.getScene().getRoot()).getChildren().add(rootContainer);
		rootContainer.setOpacity(0);
		
		final GuavacadoMainWin mainWin = this;
		progressTimeline.stop();
		progressTimeline.getKeyFrames().clear();
		progressTimeline.getKeyFrames().addAll(
				new KeyFrame(Duration.millis(popupAnim), new KeyValue(rootContainer.opacityProperty(), 1, Interpolator.EASE_OUT)),
				new KeyFrame(Duration.ZERO, new KeyValue(rootContainer.rotateProperty(), -13)),
				new KeyFrame(Duration.millis(popupAnim), new KeyValue(rootContainer.rotateProperty(), 0, Interpolator.EASE_OUT)),
				new KeyFrame(Duration.ZERO, new KeyValue(rootContainer.scaleXProperty(), 0.8)),
				new KeyFrame(Duration.ZERO, new KeyValue(rootContainer.scaleYProperty(), 0.8)),
				new KeyFrame(Duration.millis(popupAnim), new KeyValue(rootContainer.scaleXProperty(), 1, Interpolator.EASE_OUT)),
				new KeyFrame(Duration.millis(popupAnim), new KeyValue(rootContainer.scaleYProperty(), 1, Interpolator.EASE_OUT)),
				new KeyFrame(Duration.millis(popupAnim + 41), new EventHandler<ActionEvent>() {
					@Override
					public void handle(ActionEvent event) {
						switch(progressType) {
						case PROGRESSTYPE_SAVEPROJECT:
							ProjectSave.saveProject(mainWin);
							break;
						case PROGRESSTYPE_OPENPROJECT:
							ProjectOpen.openProject(mainWin, path);
							break;
						case PROGRESSTYPE_EXPORTBATCH:
							ImageExport.startBatchExport(path, lastQualitySelection, lastFormatSelection, mainWin);
							break;
						case PROGRESSTYPE_EXPORTICO:
							ImageExport.startIcoExport(path, mainWin);
							break;
						}
					}
				})
		);
		progressTimeline.play();
	}
	
	/**
	 * Updates the progress window.
	 * 
	 * @param text The text to display in the progress indicator.
	 * @param percDone Percentage done from 0-1.
	 */
	protected void updateProgress(final String text, final float percDone) {
		Pane expPane = (Pane)((Pane)((Pane)primaryStage.getScene().getRoot()).getChildren().get(((Pane)primaryStage.getScene().getRoot()).getChildren().size() - 1)).getChildren().get(0);
		((Label)((HBox)expPane.getChildren().get(2)).getChildren().get(0)).setText(text);
		((Arc)expPane.getChildren().get(1)).setLength(-360 * percDone);
	}
	
	/**
	 * Finish and animate out the progress indicator.
	 * 
	 * @param entries Images to load into the project area upon completion or null if not applicable.
	 * @param projectPath The path to the newly opened project or null if not applicable.
	 */
	protected void finishProgress(final ProjectOpen.projectImage[] entries, final File projectPath) {
		Pane rootContainer = (Pane)((Pane)primaryStage.getScene().getRoot()).getChildren().get(((Pane)primaryStage.getScene().getRoot()).getChildren().size() - 1);
		progressTimeline.getKeyFrames().clear();
		progressTimeline.getKeyFrames().addAll(
				new KeyFrame(Duration.millis(popupAnim), new KeyValue(rootContainer.opacityProperty(), 0, Interpolator.EASE_IN)),
				new KeyFrame(Duration.millis(popupAnim), new KeyValue(rootContainer.rotateProperty(), -13, Interpolator.EASE_IN)),
				new KeyFrame(Duration.millis(popupAnim), new KeyValue(rootContainer.scaleXProperty(), 0.8, Interpolator.EASE_IN)),
				new KeyFrame(Duration.millis(popupAnim), new KeyValue(rootContainer.scaleYProperty(), 0.8, Interpolator.EASE_IN)),
				new KeyFrame(Duration.millis(popupAnim + 41), new EventHandler<ActionEvent>() {
					@Override
					public void handle(ActionEvent event) {
						((Pane)primaryStage.getScene().getRoot()).getChildren().remove(((Pane)primaryStage.getScene().getRoot()).getChildren().size() - 1);
						buttonPane.setDisable(false);
						iconScrollArea.setDisable(false);
						buttonPane.requestFocus();
					}
				})
		);
		progressTimeline.play();
		
		if (entries != null) {
			if (entries.length > 0) {
				currentProject = projectPath;
				primaryStage.setTitle("Guavacado - " + currentProject.getName().substring(0, currentProject.getName().length() - 4));
				if (!tiles.isEmpty()) {
					clearProjectArea(entries);
				} else {
					beginImport(entries);
				}
			}
		}
	}
	
	/**
	 * Finish and animate the progress indicator without loading any images to the project area.
	 */
	protected void finishProgress() {
		finishProgress(null, null);
	}
	
	/**
	 * Display options dialog for batch exporting individual images.
	 */
	private void createExportDialog() {
		buttonPane.setDisable(true);
		iconScrollArea.setDisable(true);
		
		double paneWidth = 256;
		double paneHeight = 144;
		
		Pane rootContainer = new Pane();
		rootContainer.setPrefSize(paneWidth, paneHeight);
		rootContainer.setMaxSize(paneWidth, paneHeight);
		rootContainer.setMinSize(paneWidth, paneHeight);
		rootContainer.setTranslateX((primaryStage.getScene().getWidth() / 2) - (paneWidth / 2));
		rootContainer.setTranslateY((primaryStage.getScene().getHeight() / 2) - (paneHeight / 2));
		
		VBox container = new VBox();
		container.setLayoutX(0);
		container.setLayoutY(0);
		container.setPrefSize(paneWidth, paneHeight);
		container.setMaxSize(paneWidth, paneHeight);
		container.setMinSize(paneWidth, paneHeight);
		container.setAlignment(Pos.CENTER_LEFT);
		container.getStyleClass().add("popup-pane");
		rootContainer.getChildren().add(container);
		
		Label label = new Label("Select output format:");
		label.getStyleClass().add("title-lable");
		container.getChildren().add(label);
		
		final Label qualityLabel = new Label((lastFormatSelection == 1) ? "Compression:" : "N/A");
		final Slider slider = new Slider(0, 100.0, lastQualitySelection);
		final Label sliderLabel = new Label(Integer.toString((int)Math.round(lastQualitySelection)) + "%");
		sliderLabel.setMinWidth(32);
		sliderLabel.setPrefWidth(32);
		sliderLabel.setMaxWidth(32);
		if (lastFormatSelection != ImageExport.FORMAT_JPG) {
			slider.setDisable(true);
		}
		
		VBox vbox = new VBox();
		vbox.setSpacing(1);
		vbox.setFillWidth(true);
		container.setVgrow(vbox, Priority.ALWAYS);
		vbox.setAlignment(Pos.CENTER_LEFT);
		
		ChoiceBox<String> formats = new ChoiceBox<String>();
		formats.getItems().addAll("PNG", "JPG", "BMP");
		formats.getSelectionModel().select(lastFormatSelection);
		formats.getSelectionModel().selectedItemProperty().addListener(new ChangeListener<String>() {
			@Override
			public void changed(ObservableValue<? extends String> observable, String oldVal, String newVal) {
				switch (newVal) {
				case "PNG":
					lastFormatSelection = 0;
					qualityLabel.setText("N/A:");
					slider.setDisable(true);
					break;
				case "JPG":
					lastFormatSelection = 1;
					qualityLabel.setText("Quality:");
					slider.setDisable(false);
					break;
				case "BMP":
					lastFormatSelection = 2;
					qualityLabel.setText("N/A:");
					slider.setDisable(true);
					break;
				}
			}
		});
		
		HBox sliderBox = new HBox();
		sliderBox.setSpacing(5);
		sliderBox.setFillHeight(false);
		sliderBox.setAlignment(Pos.TOP_LEFT);
		sliderBox.setHgrow(slider, Priority.ALWAYS);
		sliderBox.getChildren().addAll(slider, sliderLabel);
		
		slider.valueProperty().addListener(new ChangeListener<Number>() {
			@Override
			public void changed(ObservableValue<? extends Number> observable, Number oldVal, Number newVal) {
				sliderLabel.setText(Integer.toString((int)Math.round(newVal.doubleValue())) + "%");
				lastQualitySelection = newVal.doubleValue();
			}
		});
		
		vbox.setMargin(formats, new Insets(0, 0, 3, 0));
		vbox.getChildren().addAll(formats, qualityLabel, sliderBox);
		container.getChildren().add(vbox);
		
		HBox buttonBox = new HBox();
		buttonBox.setSpacing(13);
		buttonBox.setAlignment(Pos.BOTTOM_RIGHT);
		buttonBox.setFillHeight(false);
		
		Button okay = new Button("Okay");
		okay.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent event) {
				clearExportDialog(false);
			}
		});
		Button cancel = new Button("Cancel");
		cancel.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent event) {
				clearExportDialog(true);
			}
		});
		buttonBox.getChildren().addAll(okay, cancel);
		
		container.getChildren().add(buttonBox);
		
		((Pane)primaryStage.getScene().getRoot()).getChildren().add(rootContainer);
		
		rootContainer.setOpacity(0);
		
		progressTimeline.getKeyFrames().clear();
		progressTimeline.getKeyFrames().addAll(
				new KeyFrame(Duration.millis(popupAnim), new KeyValue(rootContainer.opacityProperty(), 1, Interpolator.EASE_OUT)),
				new KeyFrame(Duration.ZERO, new KeyValue(rootContainer.rotateProperty(), -13)),
				new KeyFrame(Duration.millis(popupAnim), new KeyValue(rootContainer.rotateProperty(), 0, Interpolator.EASE_OUT)),
				new KeyFrame(Duration.ZERO, new KeyValue(rootContainer.scaleXProperty(), 0.8)),
				new KeyFrame(Duration.ZERO, new KeyValue(rootContainer.scaleYProperty(), 0.8)),
				new KeyFrame(Duration.millis(popupAnim), new KeyValue(rootContainer.scaleXProperty(), 1, Interpolator.EASE_OUT)),
				new KeyFrame(Duration.millis(popupAnim), new KeyValue(rootContainer.scaleYProperty(), 1, Interpolator.EASE_OUT))
		);
		progressTimeline.play();
	}
	
	/**
	 * Close the batch export options dialog and begin exporting images if not canceled.
	 * 
	 * @param cancel Begin batch export if false.
	 */
	private void clearExportDialog(final boolean cancel) {
		Pane rootContainer = (Pane)((Pane)primaryStage.getScene().getRoot()).getChildren().get(((Pane)primaryStage.getScene().getRoot()).getChildren().size() - 1);
		progressTimeline.getKeyFrames().clear();
		progressTimeline.getKeyFrames().addAll(
				new KeyFrame(Duration.millis(popupAnim), new KeyValue(rootContainer.opacityProperty(), 0, Interpolator.EASE_IN)),
				new KeyFrame(Duration.millis(popupAnim), new KeyValue(rootContainer.rotateProperty(), -13, Interpolator.EASE_IN)),
				new KeyFrame(Duration.millis(popupAnim), new KeyValue(rootContainer.scaleXProperty(), 0.8, Interpolator.EASE_IN)),
				new KeyFrame(Duration.millis(popupAnim), new KeyValue(rootContainer.scaleYProperty(), 0.8, Interpolator.EASE_IN)),
				new KeyFrame(Duration.millis(popupAnim + 41), new EventHandler<ActionEvent>() {
					@Override
					public void handle(ActionEvent event) {
						((Pane)primaryStage.getScene().getRoot()).getChildren().remove(((Pane)primaryStage.getScene().getRoot()).getChildren().size() - 1);
						
						if (!cancel) {
							createProgressWindow(PROGRESSTYPE_EXPORTBATCH, lastExport);
						} else {
							buttonPane.setDisable(false);
							iconScrollArea.setDisable(false);
							buttonPane.requestFocus();
						}
					}
				})
		);
		progressTimeline.play();
	}
	
	/**
	 * Center a new window on another window.
	 * 
	 * @param window The <code>Stage</code> to be centered.
	 * @param root The root element of the <code>Stage</code> to be centered.
	 * @param centerStage The <code>Stage</code> to be centered upon.
	 */
	protected void centerWindow(Stage window, Pane root, Stage centerStage) {
		GraphicsDevice gd = GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice();
		int width = gd.getDisplayMode().getWidth();
		int height = gd.getDisplayMode().getHeight();
	
		double x = ((centerStage.getWidth() / 2) + centerStage.getX()) - (root.getPrefWidth() / 2);
		double y = ((centerStage.getHeight() / 2) + centerStage.getY()) - (root.getPrefHeight() / 2);
	
		if (x < 0) {
			x = 0;
		} else if (x + root.getPrefWidth() > width) {
			x = width - root.getPrefWidth();
		}
	
		if (y < 0) {
			y = 0;
		} else if (y + root.getPrefHeight() > height) {
			y = height - root.getPrefHeight();
		}
	
		window.setX(x);
		window.setY(y);
	}
	
	/**
	 * Display an error message.
	 * 
	 * @param message The message to be displayed.
	 */
	protected void popError(final String message) {
		try {
			final Stage dialogStage = new Stage(StageStyle.DECORATED);
			
			FXMLLoader loader = new FXMLLoader(Main.class.getResource("/guavacado/resources/fxml/GuavaMessage.fxml"));
			Pane dialogPane = (Pane) loader.load();
			((GuavMessage) loader.getController()).prepare(dialogStage, "Error", message);
			
			Scene scene = new Scene(dialogPane);
			scene.getStylesheets().add(Main.class.getResource("/guavacado/resources/fxml/GuavacadoTheme.css").toExternalForm());
			dialogStage.setScene(scene);
			dialogStage.initOwner(primaryStage);
			dialogStage.initModality(Modality.APPLICATION_MODAL);
			dialogStage.getIcons().addAll(icons);
			dialogStage.setTitle("Error");
			centerWindow(dialogStage, dialogPane, primaryStage);
			dialogStage.show();
			double tmpW = scene.getWidth();
			double tmpH = scene.getHeight();
			double tmpX = dialogStage.getX();
			double tmpY = dialogStage.getY();
			dialogStage.setResizable(false);
			if (tmpW != scene.getWidth()) {
				dialogStage.setWidth(dialogStage.getWidth() - (scene.getWidth() - tmpW));
				dialogStage.setX(tmpX);
			}
			if (tmpH != scene.getHeight()) {
				dialogStage.setHeight(dialogStage.getHeight() - (scene.getHeight() - tmpH));
				dialogStage.setY(tmpY);
			}
		} catch (Exception e) {
			System.out.println("Unable to display error message due to the following error:");
			System.out.println(e.getMessage());
		}
	}
	
	/**
	 * Display a message in a popup window.
	 * 
	 * @param title Title of the popup window.
	 * @param message Message to be displayed.
	 */
	protected void popMessage(final String title, final String message) {
		try {
			final Stage dialogStage = new Stage(StageStyle.DECORATED);
			
			FXMLLoader loader = new FXMLLoader(Main.class.getResource("/guavacado/resources/fxml/GuavaMessage.fxml"));
			Pane dialogPane = (Pane) loader.load();
			((GuavMessage) loader.getController()).prepare(dialogStage, title, message);
			
			Scene scene = new Scene(dialogPane);
			scene.getStylesheets().add(Main.class.getResource("/guavacado/resources/fxml/GuavacadoTheme.css").toExternalForm());
			dialogStage.setScene(scene);
			dialogStage.initOwner(primaryStage);
			dialogStage.initModality(Modality.APPLICATION_MODAL);
			dialogStage.getIcons().addAll(icons);
			dialogStage.setTitle("FYI");
			centerWindow(dialogStage, dialogPane, primaryStage);
			dialogStage.show();
			double tmpW = scene.getWidth();
			double tmpH = scene.getHeight();
			double tmpX = dialogStage.getX();
			double tmpY = dialogStage.getY();
			dialogStage.setResizable(false);
			if (tmpW != scene.getWidth()) {
				dialogStage.setWidth(dialogStage.getWidth() - (scene.getWidth() - tmpW));
				dialogStage.setX(tmpX);
			}
			if (tmpH != scene.getHeight()) {
				dialogStage.setHeight(dialogStage.getHeight() - (scene.getHeight() - tmpH));
				dialogStage.setY(tmpY);
			}
		} catch (Exception e) {
			System.out.println("Unable to display message due to the following error:");
			System.out.println(e.getMessage());
		}
	}
	
	/**
	 * Display a dialog asking to user if they'd like to overwrite one or more files on their
	 * hard drive.
	 * 
	 * @param exportID ID of the export thread.
	 * @param files <code>List</code> of file paths that need to be overwritten.
	 */
	protected void popOverwrite(final int exportID, final List<String> files) {
		try {
			final Stage dialogStage = new Stage(StageStyle.DECORATED);
			
			FXMLLoader loader = new FXMLLoader(Main.class.getResource("/guavacado/resources/fxml/GuavaOverwrite.fxml"));
			Pane dialogPane = (Pane) loader.load();
			((GuavOverwrite) loader.getController()).prepare(exportID, dialogStage, files);
			
			Scene scene = new Scene(dialogPane);
			scene.getStylesheets().add(Main.class.getResource("/guavacado/resources/fxml/GuavacadoTheme.css").toExternalForm());
			dialogStage.setScene(scene);
			dialogStage.initOwner(primaryStage);
			dialogStage.initModality(Modality.APPLICATION_MODAL);
			dialogStage.getIcons().addAll(icons);
			dialogStage.setTitle("Overwrite");
			centerWindow(dialogStage, dialogPane, primaryStage);
			dialogStage.show();
			double tmpW = scene.getWidth();
			double tmpH = scene.getHeight();
			double tmpX = dialogStage.getX();
			double tmpY = dialogStage.getY();
			dialogStage.setResizable(false);
			if (tmpW != scene.getWidth()) {
				dialogStage.setWidth(dialogStage.getWidth() - (scene.getWidth() - tmpW));
				dialogStage.setX(tmpX);
			}
			if (tmpH != scene.getHeight()) {
				dialogStage.setHeight(dialogStage.getHeight() - (scene.getHeight() - tmpH));
				dialogStage.setY(tmpY);
			}
		} catch (Exception e) {
			System.out.println("Unable to prompt for overwrite due to the following error:");
			System.out.println(e.getMessage());
		}
	}
}
