package guavacado;

import java.awt.image.BufferedImage;
import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javafx.animation.AnimationTimer;
import javafx.animation.Interpolator;
import javafx.animation.KeyFrame;
import javafx.animation.KeyValue;
import javafx.animation.Timeline;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.embed.swing.SwingFXUtils;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.geometry.Pos;
import javafx.scene.control.Button;
import javafx.scene.control.CheckBox;
import javafx.scene.control.ChoiceBox;
import javafx.scene.control.Label;
import javafx.scene.control.TextField;
import javafx.scene.control.ToggleButton;
import javafx.scene.control.Tooltip;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.image.WritableImage;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyEvent;
import javafx.scene.layout.HBox;
import javafx.scene.layout.Pane;
import javafx.scene.layout.Priority;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.scene.shape.Arc;
import javafx.scene.shape.ArcType;
import javafx.scene.shape.Circle;
import javafx.scene.shape.Rectangle;
import javafx.scene.shape.StrokeType;
import javafx.scene.text.TextAlignment;
import javafx.util.Duration;

/**
 * An object representing an image loaded into the Guavacado project.
 * 
 * @author Adam T. Ryder http://1337atr.weebly.com
 * @version 1.0
 *
 */
public class IconTile {
	public static final Image closeTileIcon = new Image("/guavacado/resources/gfx/Button_CloseImage.png");
	public static final Image duplicateTileIcon = new Image("/guavacado/resources/gfx/Button_DuplicateImage.png");
	
	private static final double yAnimLength = Math.round(1000000000 * 0.2);
	
	public static final double width = 433;
	public static final double height = 78;
	public static final double spacing = 5;
	
	public static final int displayWidth = 64;
	public static final int displayHeight = 64;
	
	protected final File imgSrc;
	protected WritableImage wImage;
	protected Pane imagePane;
	
	protected final Label imageRes;
	protected final Label imageBDepth;
	
	protected final HBox container;
	protected final ToggleButton lockRes;
	protected final TextField scaleWidth;
	protected final TextField scaleHeight;
	protected final ChoiceBox<Integer> bitDepthSelector;
	private int bppFromProject = -1;
	protected final CheckBox pngCompress;
	private String focusedWidth = "";
	private String focusedHeight = "";
	
	protected final Button closeButton;
	protected final Button duplicateButton;
	
	private double targetX;
	private double targetY;
	private final List<Offset> yOffsets = new ArrayList<Offset>();
	
	private final Timeline timeLine = new Timeline();
	private final yAnimator yAnim = new yAnimator();
	
	private boolean loaded = false;
	private boolean error = false;
	
	private final IconTile thisTile;
	private final GuavacadoMainWin mainWin;
	
	private boolean cancel = false;
	
	protected Arc arcProgress;
	protected Label progressLabel;
	
	protected int imageWidth = 0;
	protected int imageHeight = 0;
	protected int imageBitDepth = 1;
	
	/**
	 * Constructs a new instance of <code>IconTile</code>, creates graphics and layout for the tile
	 * within the project view and begins loading the image associated with this <code>IconTile</code>.
	 * 
	 * @param imageSource The image associated with this <code>IconTile</code>.
	 * @param mainWin Guavacado main window controller class.
	 */
	protected IconTile(final File imageSource, final GuavacadoMainWin mainWin) {
		this.imgSrc = imageSource;
		this.mainWin = mainWin;
		
		container = new HBox();
		container.setPrefSize(width, height);
		container.setMaxSize(width, height);
		container.setMinSize(width, height);
		container.setSpacing(5);
		container.setAlignment(Pos.CENTER_LEFT);
		
		imagePane = new Pane();
		imagePane.setPrefSize(displayWidth, displayHeight);
		imagePane.setMaxSize(displayWidth, displayHeight);
		imagePane.setMinSize(displayWidth, displayHeight);
		container.getChildren().add(imagePane);
		
		HBox optsBox = new HBox();
		optsBox.setFillHeight(true);
		optsBox.setSpacing(5);
		container.getChildren().add(optsBox);
		container.setHgrow(optsBox, Priority.ALWAYS);
		
		VBox imageInfoBox = new VBox();
		imageInfoBox.setPrefWidth(148);
		imageInfoBox.setMaxWidth(148);
		imageInfoBox.setMinWidth(148);
		Label imageName = new Label(imgSrc.getName());
		imageName.setWrapText(false);
		imageInfoBox.getChildren().add(imageName);
		imageRes = new Label("Resolution:");
		imageRes.setWrapText(false);
		imageInfoBox.getChildren().add(imageRes);
		imageBDepth = new Label("Bit Depth:");
		imageBDepth.setWrapText(false);
		imageInfoBox.getChildren().add(imageBDepth);
		
		optsBox.getChildren().add(imageInfoBox);
		
		HBox resHContainer = new HBox();
		resHContainer.setSpacing(1);
		resHContainer.setAlignment(Pos.CENTER_LEFT);
		
		lockRes = new ToggleButton();
		lockRes.setPrefSize(15, 20);
		lockRes.setMinSize(15, 20);
		lockRes.setMaxSize(15, 20);
		lockRes.setSelected(true);
		lockRes.getStyleClass().addAll("button", "main-button", "toggle");
		lockRes.setGraphic(new ImageView(GuavacadoMainWin.avacadoToggleLockedSmall));
		lockRes.setFocusTraversable(false);
		lockRes.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent event) {
				if (lockRes.isSelected()) {
					if (!scaleWidth.getText().isEmpty()) {
						if (Integer.parseInt(scaleWidth.getText()) <= 0) {
							scaleWidth.setText("");
							scaleHeight.setText("");
						} else { scaleHeight.setText(Integer.toString(recalcHeight())); }
					} else { scaleHeight.setText(""); }
					lockRes.setGraphic(new ImageView(GuavacadoMainWin.avacadoToggleLockedSmall));
					Tooltip.install(lockRes, new Tooltip("Unlock aspect ratio"));
				} else {
					if (!scaleWidth.getText().isEmpty()) {
						if (Integer.parseInt(scaleWidth.getText()) <= 0) {
							scaleWidth.setText("");
						}
					}
					if (!scaleHeight.getText().isEmpty()) {
						if (Integer.parseInt(scaleHeight.getText()) <= 0) {
							scaleHeight.setText("");
						}
					}
					
					lockRes.setGraphic(new ImageView(GuavacadoMainWin.avacadoToggleUnlockedSmall));
					Tooltip.install(lockRes, new Tooltip("Lock aspect ratio"));
				}
			}
		});
		resHContainer.getChildren().add(lockRes);
		Tooltip.install(lockRes, new Tooltip("Unlock aspect ratio"));
		
		VBox scaleResContainer = new VBox();
		scaleResContainer.setSpacing(1);
		scaleResContainer.getChildren().add(new Label("Scale:"));
		
		scaleWidth = new TextField();
		scaleWidth.setPrefSize(48, 25);
		scaleWidth.setMaxSize(48, 25);
		scaleWidth.setMinSize(48, 25);
		scaleWidth.textProperty().addListener(new ChangeListener<String>() {
			@Override
			public void changed(ObservableValue<? extends String> observable, String oldVal, String newVal) {
				numberChecker(scaleWidth);
			}
		});
		scaleWidth.setOnKeyReleased(new EventHandler<KeyEvent>() {
			@Override
			public void handle(KeyEvent event) {
				if (event.getEventType() == KeyEvent.KEY_RELEASED && event.getCode() == KeyCode.ENTER && lockRes.isSelected()) {
					if (!scaleWidth.getText().isEmpty()) {
						if (Integer.parseInt(scaleWidth.getText()) == 0) {
							scaleWidth.setText("1");
						}
						scaleHeight.setText(Integer.toString(recalcHeight()));
					} else {
						scaleHeight.setText("");
					}
					
					focusedWidth = scaleWidth.getText();
				}
			}
		});
		scaleWidth.focusedProperty().addListener(new ChangeListener<Boolean>() {
			@Override
			public void changed(ObservableValue<? extends Boolean> observable, Boolean oldVal, Boolean newVal) {
				if (!newVal.booleanValue() && lockRes.isSelected()) {
					if (!focusedWidth.equals(scaleWidth.getText())) {
						if (!scaleWidth.getText().isEmpty()) {
							if (Integer.parseInt(scaleWidth.getText()) == 0) {
								scaleWidth.setText("1");
							}
							scaleHeight.setText(Integer.toString(recalcHeight()));
						} else {
							scaleHeight.setText("");
						}
					}
				} else if (newVal.booleanValue()) {
					focusedWidth = scaleWidth.getText();
				}
			}
		});
		Tooltip.install(scaleWidth, new Tooltip("Scale width"));
		scaleResContainer.getChildren().add(scaleWidth);
		
		scaleHeight = new TextField();
		scaleHeight.setPrefSize(48, 25);
		scaleHeight.setMaxSize(48, 25);
		scaleHeight.setMinSize(48, 25);
		scaleHeight.textProperty().addListener(new ChangeListener<String>() {
			@Override
			public void changed(ObservableValue<? extends String> observable, String oldVal, String newVal) {
				numberChecker(scaleHeight);
			}
		});
		scaleHeight.setOnKeyReleased(new EventHandler<KeyEvent>() {
			@Override
			public void handle(KeyEvent event) {
				if (event.getEventType() == KeyEvent.KEY_RELEASED && event.getCode() == KeyCode.ENTER && lockRes.isSelected()) {
					if (!scaleHeight.getText().isEmpty()) {
						if (Integer.parseInt(scaleHeight.getText()) == 0) {
							scaleHeight.setText("1");
						}
						scaleWidth.setText(Integer.toString(recalcWidth()));
					} else {
						scaleWidth.setText("");
					}
					
					focusedHeight = scaleHeight.getText();
				}
			}
		});
		scaleHeight.focusedProperty().addListener(new ChangeListener<Boolean>() {
			@Override
			public void changed(ObservableValue<? extends Boolean> observable, Boolean oldVal, Boolean newVal) {
				if (!newVal.booleanValue() && lockRes.isSelected()) {
					if (!focusedHeight.equals(scaleHeight.getText())) {
						if (!scaleHeight.getText().isEmpty()) {
							if (Integer.parseInt(scaleHeight.getText()) == 0) {
								scaleHeight.setText("1");
							}
							scaleWidth.setText(Integer.toString(recalcWidth()));
						} else {
							scaleWidth.setText("");
						}
					}
				} else if (newVal.booleanValue()) {
					focusedHeight = scaleHeight.getText();
				}
			}
		});
		Tooltip.install(scaleHeight, new Tooltip("Scale height"));
		scaleResContainer.getChildren().add(scaleHeight);
		resHContainer.getChildren().add(scaleResContainer);
		optsBox.getChildren().add(resHContainer);
		
		VBox bitDepthContainer = new VBox();
		bitDepthContainer.setSpacing(1);
		bitDepthContainer.getChildren().add(new Label("Bit Depth:"));
		
		bitDepthSelector = new ChoiceBox<Integer>();
		bitDepthSelector.setPrefSize(56, 25);
		bitDepthSelector.setMinSize(56, 25);
		bitDepthSelector.setMaxSize(56, 25);
		bitDepthContainer.getChildren().add(bitDepthSelector);
		
		pngCompress = new CheckBox("Compress");
		Tooltip.install(pngCompress, new Tooltip("Use PNG compression for .ico export"));
		bitDepthContainer.getChildren().add(pngCompress);
		optsBox.getChildren().add(bitDepthContainer);
		
		VBox buttonBox = new VBox();
		buttonBox.setAlignment(Pos.CENTER_RIGHT);
		buttonBox.setSpacing(3);
		closeButton = new Button();
		closeButton.setGraphic(new ImageView(closeTileIcon));
		closeButton.setFocusTraversable(false);
		closeButton.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent event) {
				mainWin.removeItem(thisTile);
			}
		});
		closeButton.getStyleClass().addAll("button", "main-button", "toggle");
		Tooltip.install(closeButton, new Tooltip("Remove image from project"));
		buttonBox.getChildren().add(closeButton);
		
		duplicateButton = new Button();
		duplicateButton.setGraphic(new ImageView(duplicateTileIcon));
		duplicateButton.setFocusTraversable(false);
		duplicateButton.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent event) {
				mainWin.cloneItem(thisTile);
			}
		});
		duplicateButton.getStyleClass().addAll("button", "main-button", "toggle");
		Tooltip.install(duplicateButton, new Tooltip("Duplicate image"));
		buttonBox.getChildren().add(duplicateButton);
		optsBox.getChildren().add(buttonBox);
		optsBox.setHgrow(buttonBox, Priority.ALWAYS);
		
		container.getStyleClass().add("icon-tile");
		container.setDisable(true);
		thisTile = this;
		
		createProgressIndicator();
		ImageLoader.addImage(this);
	}
	
	/**
	 * Constructs a new instance of <code>IconTile</code>, creates graphics and layout for the tile
	 * within the project view and begins.  Used to clone an already loaded <code>IconTile</code>.
	 * 
	 * @param tile The <code>IconTile</code> to clone from.
	 * @param listIndex Index of the <code>IconTile</code> to be cloned within the project
	 * <code>List</code> of <code>IconTile</code>s.
	 * @param parent Element within the main Guavacado Window to which this <code>IconTile</code> will be
	 * contained.
	 * @param mainWin Guavacado main window controller class.
	 */
	protected IconTile(final IconTile tile, final int listIndex, final Pane parent, final GuavacadoMainWin mainWin) {
		imgSrc = new File(tile.imgSrc.getPath());
		this.mainWin = mainWin;
		loaded = true;
		
		imageWidth = tile.imageWidth;
		imageHeight = tile.imageHeight;
		imageBitDepth = tile.imageBitDepth;
		
		container = new HBox();
		container.setPrefSize(width, height);
		container.setMaxSize(width, height);
		container.setMinSize(width, height);
		container.setSpacing(5);
		container.setAlignment(Pos.CENTER_LEFT);
		
		imagePane = new Pane();
		imagePane.setPrefSize(displayWidth, displayHeight);
		imagePane.setMaxSize(displayWidth, displayHeight);
		imagePane.setMinSize(displayWidth, displayHeight);
		container.getChildren().add(imagePane);
		
		if (!tile.error) {
			wImage = tile.wImage;
			ImageView iView = new ImageView(wImage);
			if (wImage.getWidth() > displayWidth && wImage.getHeight() > displayHeight) {
				if (wImage.getWidth() > wImage.getHeight()) {
					iView.setFitWidth(displayWidth);
				} else {
					iView.setFitHeight(displayHeight);
				}
				iView.setSmooth(true);
			} else if (wImage.getHeight() > displayHeight) {
				iView.setFitHeight(displayHeight);
				iView.setSmooth(true);
			} else if (wImage.getWidth() > displayWidth) {
				iView.setFitWidth(displayWidth);
				iView.setSmooth(true);
			} else if (wImage.getWidth() > wImage.getHeight()) {
				iView.setFitWidth(wImage.getWidth());
			} else {
				iView.setFitHeight(wImage.getHeight());
			}
			iView.setPreserveRatio(true);
			iView.setCache(false);
			
			if (iView.getBoundsInLocal().getWidth() < displayWidth) {
				iView.setTranslateX((displayWidth / 2) - (iView.getBoundsInLocal().getWidth() / 2));
			}
			
			if (iView.getBoundsInLocal().getHeight() < displayHeight) {
				iView.setTranslateY((displayHeight / 2) - (iView.getBoundsInLocal().getHeight() / 2));
			}
			imagePane.getChildren().add(iView);
		} else {
			error = true;
			Rectangle rectangle = new Rectangle(0, 0, displayWidth, displayHeight);
			rectangle.setArcWidth(11);
			rectangle.setArcHeight(11);
			rectangle.setFill(Color.LIGHTGRAY);
			rectangle.setStrokeWidth(0);
			
			HBox textBox = new HBox();
			textBox.setPrefSize(displayWidth, displayHeight);
			textBox.setMinSize(displayWidth, displayHeight);
			textBox.setMaxSize(displayWidth, displayHeight);
			textBox.setStyle("-fx-padding: 3;");
			textBox.setAlignment(Pos.CENTER);
			
			progressLabel = new Label("Out of Memory!");
			progressLabel.setStyle("-fx-text-fill: rgb(80, 80, 80);"
					+ "-fx-font-size: 12px;");
			progressLabel.setWrapText(true);
			progressLabel.setTextAlignment(TextAlignment.CENTER);
			textBox.getChildren().add(progressLabel);
			
			imagePane.getChildren().clear();
			imagePane.getChildren().addAll(rectangle, textBox);
		}
		
		HBox optsBox = new HBox();
		optsBox.setFillHeight(true);
		optsBox.setSpacing(5);
		container.getChildren().add(optsBox);
		container.setHgrow(optsBox, Priority.ALWAYS);
		
		VBox imageInfoBox = new VBox();
		imageInfoBox.setPrefWidth(148);
		imageInfoBox.setMaxWidth(148);
		imageInfoBox.setMinWidth(148);
		Label imageName = new Label(imgSrc.getName());
		imageName.setWrapText(false);
		imageInfoBox.getChildren().add(imageName);
		imageRes = new Label(tile.imageRes.getText());
		imageRes.setWrapText(false);
		imageInfoBox.getChildren().add(imageRes);
		imageBDepth = new Label(tile.imageBDepth.getText());
		imageBDepth.setWrapText(false);
		imageInfoBox.getChildren().add(imageBDepth);
		
		optsBox.getChildren().add(imageInfoBox);
		
		HBox resHContainer = new HBox();
		resHContainer.setSpacing(1);
		resHContainer.setAlignment(Pos.CENTER_LEFT);
		
		lockRes = new ToggleButton();
		lockRes.setPrefSize(15, 20);
		lockRes.setMinSize(15, 20);
		lockRes.setMaxSize(15, 20);
		lockRes.getStyleClass().addAll("button", "main-button", "toggle");
		if (tile.lockRes.isSelected()) {
			lockRes.setSelected(true);
			lockRes.setGraphic(new ImageView(GuavacadoMainWin.avacadoToggleLockedSmall));
			Tooltip.install(lockRes, new Tooltip("Unlock aspect ratio"));
		} else {
			lockRes.setGraphic(new ImageView(GuavacadoMainWin.avacadoToggleUnlockedSmall));
			Tooltip.install(lockRes, new Tooltip("Lock aspect ratio"));
		}
		lockRes.setFocusTraversable(false);
		lockRes.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent event) {
				if (lockRes.isSelected()) {
					if (!scaleWidth.getText().isEmpty()) {
						if (Integer.parseInt(scaleWidth.getText()) <= 0) {
							scaleWidth.setText("");
							scaleHeight.setText("");
						} else { scaleHeight.setText(Integer.toString(recalcHeight())); }
					} else { scaleHeight.setText(""); }
					lockRes.setGraphic(new ImageView(GuavacadoMainWin.avacadoToggleLockedSmall));
					Tooltip.install(lockRes, new Tooltip("Unlock aspect ratio"));
				} else {
					if (!scaleWidth.getText().isEmpty()) {
						if (Integer.parseInt(scaleWidth.getText()) <= 0) {
							scaleWidth.setText("");
						}
					}
					if (!scaleHeight.getText().isEmpty()) {
						if (Integer.parseInt(scaleHeight.getText()) <= 0) {
							scaleHeight.setText("");
						}
					}
					
					lockRes.setGraphic(new ImageView(GuavacadoMainWin.avacadoToggleUnlockedSmall));
					Tooltip.install(lockRes, new Tooltip("Lock aspect ratio"));
				}
			}
		});
		resHContainer.getChildren().add(lockRes);
		
		VBox scaleResContainer = new VBox();
		scaleResContainer.setSpacing(1);
		scaleResContainer.getChildren().add(new Label("Scale:"));
		
		scaleWidth = new TextField();
		scaleWidth.setPrefSize(48, 25);
		scaleWidth.setMaxSize(48, 25);
		scaleWidth.setMinSize(48, 25);
		scaleWidth.setText(tile.scaleWidth.getText());
		scaleWidth.textProperty().addListener(new ChangeListener<String>() {
			@Override
			public void changed(ObservableValue<? extends String> observable, String oldVal, String newVal) {
				numberChecker(scaleWidth);
			}
		});
		scaleWidth.setOnKeyReleased(new EventHandler<KeyEvent>() {
			@Override
			public void handle(KeyEvent event) {
				if (event.getEventType() == KeyEvent.KEY_RELEASED && event.getCode() == KeyCode.ENTER && lockRes.isSelected()) {
					if (!scaleWidth.getText().isEmpty()) {
						if (Integer.parseInt(scaleWidth.getText()) == 0) {
							scaleWidth.setText("1");
						}
						scaleHeight.setText(Integer.toString(recalcHeight()));
					} else {
						scaleHeight.setText("");
					}
					
					focusedWidth = scaleWidth.getText();
				}
			}
		});
		scaleWidth.focusedProperty().addListener(new ChangeListener<Boolean>() {
			@Override
			public void changed(ObservableValue<? extends Boolean> observable, Boolean oldVal, Boolean newVal) {
				if (!newVal.booleanValue() && lockRes.isSelected()) {
					if (!focusedWidth.equals(scaleWidth.getText())) {
						if (!scaleWidth.getText().isEmpty()) {
							if (Integer.parseInt(scaleWidth.getText()) == 0) {
								scaleWidth.setText("1");
							}
							scaleHeight.setText(Integer.toString(recalcHeight()));
						} else {
							scaleHeight.setText("");
						}
					}
				} else if (newVal.booleanValue()) {
					focusedWidth = scaleWidth.getText();
				}
			}
		});
		Tooltip.install(scaleWidth, new Tooltip("Scale width"));
		scaleResContainer.getChildren().add(scaleWidth);
		
		scaleHeight = new TextField();
		scaleHeight.setPrefSize(48, 25);
		scaleHeight.setMaxSize(48, 25);
		scaleHeight.setMinSize(48, 25);
		scaleHeight.setText(tile.scaleHeight.getText());
		scaleHeight.textProperty().addListener(new ChangeListener<String>() {
			@Override
			public void changed(ObservableValue<? extends String> observable, String oldVal, String newVal) {
				numberChecker(scaleHeight);
			}
		});
		scaleHeight.setOnKeyReleased(new EventHandler<KeyEvent>() {
			@Override
			public void handle(KeyEvent event) {
				if (event.getEventType() == KeyEvent.KEY_RELEASED && event.getCode() == KeyCode.ENTER && lockRes.isSelected()) {
					if (!scaleHeight.getText().isEmpty()) {
						if (Integer.parseInt(scaleHeight.getText()) == 0) {
							scaleHeight.setText("1");
						}
						scaleWidth.setText(Integer.toString(recalcWidth()));
					} else {
						scaleWidth.setText("");
					}
					
					focusedHeight = scaleHeight.getText();
				}
			}
		});
		scaleHeight.focusedProperty().addListener(new ChangeListener<Boolean>() {
			@Override
			public void changed(ObservableValue<? extends Boolean> observable, Boolean oldVal, Boolean newVal) {
				if (!newVal.booleanValue() && lockRes.isSelected()) {
					if (!focusedHeight.equals(scaleHeight.getText())) {
						if (!scaleHeight.getText().isEmpty()) {
							if (Integer.parseInt(scaleHeight.getText()) == 0) {
								scaleHeight.setText("1");
							}
							scaleWidth.setText(Integer.toString(recalcWidth()));
						} else {
							scaleWidth.setText("");
						}
					}
				} else if (newVal.booleanValue()) {
					focusedHeight = scaleHeight.getText();
				}
			}
		});
		Tooltip.install(scaleHeight, new Tooltip("Scale height"));
		scaleResContainer.getChildren().add(scaleHeight);
		resHContainer.getChildren().add(scaleResContainer);
		optsBox.getChildren().add(resHContainer);
		
		VBox bitDepthContainer = new VBox();
		bitDepthContainer.setSpacing(1);
		bitDepthContainer.getChildren().add(new Label("Bit Depth:"));
		
		bitDepthSelector = new ChoiceBox<Integer>();
		bitDepthSelector.setPrefSize(56, 25);
		bitDepthSelector.setMinSize(56, 25);
		bitDepthSelector.setMaxSize(56, 25);
		bitDepthSelector.getItems().addAll(tile.bitDepthSelector.getItems());
		bitDepthSelector.getSelectionModel().select(tile.bitDepthSelector.getSelectionModel().getSelectedIndex());
		bitDepthContainer.getChildren().add(bitDepthSelector);
		
		pngCompress = new CheckBox("Compress");
		pngCompress.setSelected(tile.pngCompress.isSelected());
		Tooltip.install(pngCompress, new Tooltip("Use PNG compression for .ico export"));
		bitDepthContainer.getChildren().add(pngCompress);
		optsBox.getChildren().add(bitDepthContainer);
		
		VBox buttonBox = new VBox();
		buttonBox.setAlignment(Pos.CENTER_RIGHT);
		buttonBox.setSpacing(3);
		closeButton = new Button();
		closeButton.setGraphic(new ImageView(closeTileIcon));
		closeButton.setFocusTraversable(false);
		closeButton.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent event) {
				mainWin.removeItem(thisTile);
			}
		});
		closeButton.getStyleClass().addAll("button", "main-button", "toggle");
		Tooltip.install(closeButton, new Tooltip("Remove image from project"));
		buttonBox.getChildren().add(closeButton);
		
		duplicateButton = new Button();
		duplicateButton.setGraphic(new ImageView(duplicateTileIcon));
		duplicateButton.setFocusTraversable(false);
		duplicateButton.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent event) {
				mainWin.cloneItem(thisTile);
			}
		});
		duplicateButton.getStyleClass().addAll("button", "main-button", "toggle");
		Tooltip.install(duplicateButton, new Tooltip("Duplicate image"));
		buttonBox.getChildren().add(duplicateButton);
		optsBox.getChildren().add(buttonBox);
		optsBox.setHgrow(buttonBox, Priority.ALWAYS);
		
		container.getStyleClass().add("icon-tile");
		container.setDisable(true);
		thisTile = this;
		
		targetY = 24 + (listIndex * (height + spacing));
		targetX = 22;
		container.setTranslateY(tile.container.getTranslateY());
		container.setTranslateX(targetX);
		
		parent.getChildren().add(container);
		container.toFront();
		for (int i = listIndex + 1; i < mainWin.getTiles().size(); i++) {
			mainWin.getTiles().get(i).container.toFront();
		}
		
		timeLine.getKeyFrames().addAll(
				new KeyFrame(Duration.ZERO, new KeyValue(container.opacityProperty(), 0)),
				new KeyFrame(Duration.millis(200), new KeyValue(container.opacityProperty(), 1, Interpolator.EASE_BOTH)),
				new KeyFrame(Duration.millis(200), new EventHandler<ActionEvent>() {
					@Override
					public void handle(ActionEvent event) {
						container.setDisable(false);
						container.opacityProperty().addListener(new ChangeListener<Number>() {
							@Override
							public void changed(ObservableValue<? extends Number> observable, Number oldVal, Number newVal) {
								if (newVal.doubleValue() < 1) {
									container.setOpacity(1);
								}
							}
						});
					}
				})
		);
		timeLine.setCycleCount(1);
		timeLine.play();
		
		setNewYOffset(-(IconTile.height + IconTile.spacing));
	}
	
	/**
	 * Checks the text content of a <code>TextField</code> to ensure that only integer values
	 * are entered.
	 * 
	 * @param field The <code>TextField</code> to check.
	 */
	private void numberChecker(TextField field) {
		String s = field.getText();
		whileLoop: while (true) {
			for (int i = 0; i < s.length(); i++) {
				char c = s.charAt(i);
				if (!(c >= '0' && c <= '9')) {
					if (i > 0 && i < s.length() - 1) {
						s = s.substring(0, i) + s.substring(i + 1, s.length());
					} else if (i > 0) {
						s = s.substring(0, i);
					} else if (s.length() > 1){
						s = s.substring(1, s.length());
					} else {
						s = "";
					}
					continue whileLoop;
				}
			}
			
			break;
		}
		
		if (!s.isEmpty()) {
			try {
				int i = Integer.parseInt(s);
				if (i <= 0) {
					s = "";
				}
			} catch (Exception e) {
				
			}
		}
		
		field.setText(s);
	}
	
	/**
	 * Set the content of the tile to display information from a tile loaded
	 * from a saved project.
	 * 
	 * @param pi the <code>ProjectOpen.projectImage</code> to pull the saved
	 * settings from.
	 * @see ProjectOpen.projectImage
	 */
	protected void setFromProjectLoad(ProjectOpen.projectImage pi) {
		if (pi.width > 0) { scaleWidth.setText(Integer.toString(pi.width)); }
		if (pi.height > 0) { scaleHeight.setText(Integer.toString(pi.height)); }
		bppFromProject = pi.bpp;
		pngCompress.setSelected(pi.compress);
		lockRes.setSelected(pi.isResLocked);
		lockRes.setGraphic(new ImageView((lockRes.isSelected()) ? GuavacadoMainWin.avacadoToggleLockedSmall : GuavacadoMainWin.avacadoToggleUnlockedSmall));
	}
	
	/**
	 * Recalculates the width setting for scaling the image.  Used when the height
	 * value is changed and the lock aspect ratio setting is on.
	 * 
	 * @return The new width.
	 */
	private int recalcWidth() {
		double ratio = imageWidth / (double)imageHeight;
		int newWidth = (int)Math.round(Integer.parseInt(scaleHeight.getText()) * ratio);
		return (newWidth > 0) ? newWidth : 1;
	}
	
	/**
	 * Recalculates the height setting for scaling the image.  Used when the width
	 * value is changed and the lock aspect ratio setting is on.
	 * 
	 * @return The new height.
	 */
	private int recalcHeight() {
		double ratio = imageHeight / (double)imageWidth;
		int newHeight = (int)Math.round(Integer.parseInt(scaleWidth.getText()) * ratio);
		return (newHeight > 0) ? newHeight : 1;
	}
	
	/**
	 * Adds the image tile to the project window.  Tile will animate in from
	 * right to left.
	 * 
	 * @param listIndex The index of the tile within the project's <code>ImageTile</code>
	 * <code>List</code>.
	 * @param parent The <code>Pane that will display the tile.</code>
	 */
	protected void addToScene(final int listIndex, final Pane parent) {
		targetY = 24 + (listIndex * (height + spacing));
		container.setTranslateY(targetY);
		targetX = 22;
		container.setTranslateX(targetX);
		
		parent.getChildren().add(container);
		container.toFront();
		
		timeLine.getKeyFrames().addAll(
				new KeyFrame(Duration.ZERO, new KeyValue(container.translateXProperty(), parent.getWidth())),
				new KeyFrame(Duration.millis(167), new KeyValue(container.translateXProperty(), targetX, Interpolator.EASE_OUT)),
				new KeyFrame(Duration.millis(167), new EventHandler<ActionEvent>() {
					@Override
					public void handle(ActionEvent event) {
						if (loaded) {
							container.setDisable(false);
						}
					}
				})
		);
		timeLine.setCycleCount(1);
		timeLine.play();
	}
	
	/**
	 * 
	 * @return The width scale setting of the image or the images width if no
	 * scale has been entered.
	 */
	public int getScaledWidth() {
		return (scaleWidth.getText().isEmpty()) ? imageWidth : Integer.parseInt(scaleWidth.getText());
	}
	
	/**
	 * 
	 * @return The height scale setting of the image or the images height if no
	 * scale has been entered.
	 */
	public int getScaledHeight() {
		return (scaleHeight.getText().isEmpty()) ? imageHeight : Integer.parseInt(scaleHeight.getText());
	}
	
	/**
	 * 
	 * @return The selected bit-depth of the image.
	 */
	public int getOutputBpp() {
		return bitDepthSelector.getSelectionModel().getSelectedItem().intValue();
	}
	
	/**
	 * 
	 * @return The selected bit-depth of the image adjusted for .ico output.
	 */
	public int getOutputBppICO() {
		int bpp = bitDepthSelector.getSelectionModel().getSelectedItem().intValue();
		return (bpp == 16) ? 24 : bpp;
	}
	
	/**
	 * 
	 * @return True if the user has opted to use PNG compression for .ico output.
	 */
	public boolean pngOutput() {
		return pngCompress.isSelected();
	}
	
	/**
	 * Creates a progress indicator display in place of the image thumbnail.
	 * Used when the tile has been added to the scene, but the image is still
	 * loading.
	 */
	protected void createProgressIndicator() {
		Rectangle rectangle = new Rectangle(0, 0, displayWidth, displayHeight);
		rectangle.setArcWidth(11);
		rectangle.setArcHeight(11);
		rectangle.setFill(Color.LIGHTGRAY);
		rectangle.setStrokeWidth(0);
		
		Circle arcBackground = new Circle(displayWidth / 2f, displayHeight / 2f, (displayWidth / 2f) * 0.8f, Color.TRANSPARENT);
		arcBackground.setStroke(Color.DARKGREY);
		arcBackground.setStrokeWidth(7);
		arcBackground.setStrokeType(StrokeType.CENTERED);
		
		arcProgress = new Arc(displayWidth / 2f, displayHeight / 2f, (displayWidth / 2f) * 0.8f, (displayHeight / 2f) * 0.8f, 90, 0);
		arcProgress.setType(ArcType.OPEN);
		arcProgress.setFill(Color.TRANSPARENT);
		arcProgress.setStroke(Color.CYAN);
		arcProgress.setStrokeWidth(5);
		arcProgress.setStrokeType(StrokeType.CENTERED);
		
		HBox textBox = new HBox();
		textBox.setPrefSize(displayWidth, displayHeight);
		textBox.setMinSize(displayWidth, displayHeight);
		textBox.setMaxSize(displayWidth, displayHeight);
		textBox.setAlignment(Pos.CENTER);
		
		progressLabel = new Label("0%");
		progressLabel.setStyle("-fx-text-fill: rgb(80, 80, 80);"
				+ "-fx-font-size: 12px;");
		textBox.getChildren().add(progressLabel);
		
		imagePane.getChildren().clear();
		imagePane.getChildren().addAll(rectangle, arcBackground, arcProgress, textBox);
	}
	
	/**
	 * Displays an out of memory message in place of the image thumbnail if the
	 * thumbnail could not be displayed due to a lack of system memory.
	 * 
	 * @param width The width of the image.
	 * @param height The height of the image.
	 * @param bitDepth The bit-depth of the image.
	 */
	protected void setOutOfMemory(final int width, final int height, int bitDepth) {
		error = true;
		
		imageWidth = width;
		imageHeight = height;
		imageBitDepth = bitDepth;
		
		Rectangle rectangle = new Rectangle(0, 0, displayWidth, displayHeight);
		rectangle.setArcWidth(11);
		rectangle.setArcHeight(11);
		rectangle.setFill(Color.LIGHTGRAY);
		rectangle.setStrokeWidth(0);
		
		HBox textBox = new HBox();
		textBox.setPrefSize(displayWidth, displayHeight);
		textBox.setMinSize(displayWidth, displayHeight);
		textBox.setMaxSize(displayWidth, displayHeight);
		textBox.setStyle("-fx-padding: 3;");
		textBox.setAlignment(Pos.CENTER);
		
		progressLabel = new Label("Out of Memory!");
		progressLabel.setStyle("-fx-text-fill: rgb(80, 80, 80);"
				+ "-fx-font-size: 12px;");
		progressLabel.setWrapText(true);
		progressLabel.setTextAlignment(TextAlignment.CENTER);
		textBox.getChildren().add(progressLabel);
		
		imagePane.getChildren().clear();
		imagePane.getChildren().addAll(rectangle, textBox);
		
		if (!cancel) {
			container.setDisable(false);
			loaded = true;
			duplicateButton.setDisable(true);
			scaleHeight.setDisable(true);
			scaleWidth.setDisable(true);
			bitDepthSelector.setDisable(true);
			pngCompress.setDisable(true);
			
			imageRes.setText("Resolution:  " + Integer.toString(width) + "x" + Integer.toString(height));
			imageBDepth.setText("Bit Depth:  " + Integer.toString(bitDepth));
			if (bitDepth >= 1) {
				bitDepthSelector.getItems().add(new Integer(1));
			}
			if (bitDepth >= 4) {
				bitDepthSelector.getItems().add(new Integer(4));
			}
			if (bitDepth >= 8) {
				bitDepthSelector.getItems().add(new Integer(8));
			}
			if (bitDepth >= 16) {
				bitDepthSelector.getItems().add(new Integer(16));
			}
			if (bitDepth >= 24) {
				bitDepthSelector.getItems().add(new Integer(24));
			}
			if (bitDepth >= 32) {
				bitDepthSelector.getItems().add(new Integer(32));
			}
			
			if (bppFromProject >= 1 && bppFromProject <= bitDepth) {
				switch(bppFromProject) {
				case 1:
					bitDepthSelector.getSelectionModel().select(0);
					break;
				case 4:
					bitDepthSelector.getSelectionModel().select(1);
					break;
				case 8:
					bitDepthSelector.getSelectionModel().select(2);
					break;
				case 16:
					bitDepthSelector.getSelectionModel().select(3);
					break;
				case 24:
					bitDepthSelector.getSelectionModel().select(4);
					break;
				default:
					bitDepthSelector.getSelectionModel().select(5);
					break;
				}
			} else {
				bitDepthSelector.getSelectionModel().select(bitDepthSelector.getItems().size() - 1);
			}
		}
	}
	
	/**
	 * 
	 * @return True if this image tile errored while loading such as having run
	 * out of memory while loading the image.
	 */
	public boolean isError() {
		return error;
	}
	
	/**
	 * Display an error message in place of the image thumbnail if an error was
	 * encountered while loading the image.
	 */
	protected void setError() {
		error = true;
		
		Rectangle rectangle = new Rectangle(0, 0, displayWidth, displayHeight);
		rectangle.setArcWidth(11);
		rectangle.setArcHeight(11);
		rectangle.setFill(Color.LIGHTGRAY);
		rectangle.setStrokeWidth(0);
		
		HBox textBox = new HBox();
		textBox.setPrefSize(displayWidth, displayHeight);
		textBox.setMinSize(displayWidth, displayHeight);
		textBox.setMaxSize(displayWidth, displayHeight);
		textBox.setStyle("-fx-padding: 3;");
		textBox.setAlignment(Pos.CENTER);
		
		progressLabel = new Label("Error Loading");
		progressLabel.setStyle("-fx-text-fill: rgb(80, 80, 80);"
				+ "-fx-font-size: 12px;");
		progressLabel.setWrapText(true);
		progressLabel.setTextAlignment(TextAlignment.CENTER);
		textBox.getChildren().add(progressLabel);
		
		imagePane.getChildren().clear();
		imagePane.getChildren().addAll(rectangle, textBox);
		
		if (!cancel) {
			container.setDisable(false);
			loaded = true;
			duplicateButton.setDisable(true);
			scaleHeight.setDisable(true);
			scaleWidth.setDisable(true);
			bitDepthSelector.setDisable(true);
			pngCompress.setDisable(true);
			
			imageRes.setText("Resolution:  ?x?");
			imageBDepth.setText("Bit Depth:  ?");
		}
	}
	
	/**
	 * Set the image thumbnail and attributes to be displayed in the <code>ImageTile</code>
	 * upon this information becoming available.
	 * 
	 * @param image The <code>BufferedImage</code> thumbnail.
	 * @param width Width of the original image in pixels.
	 * @param height Height of the original image in pixels.
	 * @param bitDepth Bit-depth of the original image.
	 */
	protected void setImage(final BufferedImage image, final int width, final int height, int bitDepth) {
		loaded = true;
		imageWidth = width;
		imageHeight = height;
		imageBitDepth = bitDepth;
		
		wImage = SwingFXUtils.toFXImage(image, null);
		imagePane.getChildren().clear();
		ImageView iView = new ImageView(wImage);
		if (wImage.getWidth() > displayWidth && wImage.getHeight() > displayHeight) {
			if (wImage.getWidth() > wImage.getHeight()) {
				iView.setFitWidth(displayWidth);
			} else {
				iView.setFitHeight(displayHeight);
			}
			iView.setSmooth(true);
		} else if (wImage.getHeight() > displayHeight) {
			iView.setFitHeight(displayHeight);
			iView.setSmooth(true);
		} else if (wImage.getWidth() > displayWidth) {
			iView.setFitWidth(displayWidth);
			iView.setSmooth(true);
		} else if (wImage.getWidth() > wImage.getHeight()) {
			iView.setFitWidth(wImage.getWidth());
		} else {
			iView.setFitHeight(wImage.getHeight());
		}
		iView.setPreserveRatio(true);
		iView.setCache(false);
		
		if (iView.getBoundsInLocal().getWidth() < displayWidth) {
			iView.setTranslateX((displayWidth / 2f) - (iView.getBoundsInLocal().getWidth() / 2));
		}
		
		if (iView.getBoundsInLocal().getHeight() < displayHeight) {
			iView.setTranslateY((displayHeight / 2) - (iView.getBoundsInLocal().getHeight() / 2));
		}
		
		imagePane.getChildren().add(iView);
		
		imageRes.setText("Resolution:  " + Integer.toString(width) + "x" + Integer.toString(height));
		imageBDepth.setText("Bit Depth:  " + Integer.toString(bitDepth));
		
		if (bitDepth >= 1) {
			bitDepthSelector.getItems().add(new Integer(1));
		}
		if (bitDepth >= 4) {
			bitDepthSelector.getItems().add(new Integer(4));
		}
		if (bitDepth >= 8) {
			bitDepthSelector.getItems().add(new Integer(8));
		}
		if (bitDepth >= 16) {
			bitDepthSelector.getItems().add(new Integer(16));
		}
		if (bitDepth >= 24) {
			bitDepthSelector.getItems().add(new Integer(24));
		}
		if (bitDepth >= 32) {
			bitDepthSelector.getItems().add(new Integer(32));
		}
		if (bppFromProject >= 1 && bppFromProject <= bitDepth) {
			switch(bppFromProject) {
			case 1:
				bitDepthSelector.getSelectionModel().select(0);
				break;
			case 4:
				bitDepthSelector.getSelectionModel().select(1);
				break;
			case 8:
				bitDepthSelector.getSelectionModel().select(2);
				break;
			case 16:
				bitDepthSelector.getSelectionModel().select(3);
				break;
			case 24:
				bitDepthSelector.getSelectionModel().select(4);
				break;
			default:
				bitDepthSelector.getSelectionModel().select(5);
				break;
			}
		} else {
			bitDepthSelector.getSelectionModel().select(bitDepthSelector.getItems().size() - 1);
		}
		
		if (!cancel) {
			container.setDisable(false);
		}
	}
	
	/**
	 * Removes an <code>ImageTile</code> from the project view. The tile will animate
	 * out from left to right and then be removed.
	 */
	protected void removeFromScene() {
		container.toFront();
		container.setDisable(true);
		timeLine.stop();
		timeLine.getKeyFrames().clear();
		timeLine.getKeyFrames().addAll(
				new KeyFrame(Duration.ZERO, new KeyValue(container.translateXProperty(), targetX)),
				new KeyFrame(Duration.millis(84), new KeyValue(container.translateXProperty(), targetX - 20, Interpolator.EASE_BOTH)),
				new KeyFrame(Duration.millis(167 + 84), new KeyValue(container.translateXProperty(), ((Pane)container.getParent()).getWidth(), Interpolator.EASE_IN)),
				new KeyFrame(Duration.millis(167 + 84), new EventHandler<ActionEvent>() {
					@Override
					public void handle(ActionEvent event) {
						if (!loaded) {
							stopOperations();
						}
						((Pane)container.getParent()).getChildren().remove(container);
					}
				})
		);
		timeLine.play();
	}
	
	/**
	 * Sets the cancel flag to true and stops all animations on this <code>ImageTile</code>.
	 */
	protected void stopOperations() {
		cancel = true;
		yAnim.stop();
		timeLine.stop();
	}
	
	/**
	 * 
	 * @return True if this <code>ImageTile</code> has been cancelled.
	 */
	public boolean isCancelled() {
		return cancel;
	}
	
	/**
	 * 
	 * @return True if this <code>ImageTile</code>'s thumbnail and information
	 * has been loaded.
	 */
	public boolean isLoaded() {
		return loaded;
	}
	
	/**
	 * Adds an additional <code>ImageTile.Offset</code> to this tile's <code>List</code>
	 * of offsets on the Y axis.  Used when adding, cloning and deleting
	 * <code>ImageTile</code>s.
	 * 
	 * @param offset The new offset to add. Typically the height of an
	 * <code>ImageTile</code> plus margin.
	 * @see ImageTile.Offset
	 * @see ImageTile.yAnimator
	 */
	protected void setNewYOffset(final double offset) {
		targetY -= offset;
		yOffsets.add(new Offset(offset));
		if (!yAnim.isPlaying) {
			yAnim.start();
		}
	}
	
	/**
	 * Stores information in regards to animating an <code>ImageTile</code> on
	 * the Y axis.  Multiple offsets are stored in a <code>List</code> and
	 * iterated over so an <code>ImageTile</code> can keep track of where it
	 * will ultimately end up when animating after an <code>ImageTile</code>
	 * has been added or removed from the scene.
	 * 
	 * @author Adam T. Ryder
	 * @see ImageTile.yAnimator
	 *
	 */
	private class Offset {
		private final double offset;
		private long lastFrame = 0;
		private long animTme = 0;
		
		private Offset(final double offset) {
			this.offset = offset;
			lastFrame = System.nanoTime();
		}
	}
	
	/**
	 * An <code>AnimationTimer</code> used to animate the <code>ImageTile</code>
	 * on the Y axis when <code>ImageTile</code>s above have been added or
	 * removed from the scene.
	 * 
	 * Iterates over this <code>ImageTile</code>s list of <code>Offset</code>s
	 * adding that <code>Offset</code>s interpolated value to the current offset.
	 * 
	 * @author Adam T. Ryder
	 * @see ImageTile.Offset
	 *
	 */
	private class yAnimator extends AnimationTimer {
		private boolean isPlaying = false;
		
		@Override
		public void handle(long now) {
			double currentOffset = 0;
			for (Iterator<Offset> it = yOffsets.iterator(); it.hasNext();) {
				Offset o = it.next();
				o.animTme += now - o.lastFrame;
				currentOffset += Interpolator.EASE_BOTH.interpolate(o.offset, 0, Math.min(1, o.animTme / yAnimLength));
				if (o.animTme >= yAnimLength) {
					it.remove();
				} else {
					o.lastFrame = now;
				}
			}
			container.setTranslateY(targetY + currentOffset);
			
			if (mainWin.getTiles().indexOf(thisTile) == mainWin.getTiles().size() - 1) {
				mainWin.getListPane().setMaxHeight(container.getTranslateY() + 24 + height + spacing);
				mainWin.getListPane().setMinHeight(container.getTranslateY() + 24 + height + spacing);
				mainWin.getListPane().setPrefHeight(container.getTranslateY() + 24 + height + spacing);
			}
			
			if (yOffsets.isEmpty()) {
				this.stop();
			}
		}
		
		@Override
		public void start() {
			isPlaying = true;
			super.start();
		}
		
		@Override
		public void stop() {
			isPlaying = false;
			super.stop();
		}
	}
}
