package guavacado;

import java.awt.AlphaComposite;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.awt.image.ColorConvertOp;
import java.awt.image.IndexColorModel;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javafx.application.Platform;

import javax.imageio.IIOImage;
import javax.imageio.ImageIO;
import javax.imageio.ImageReader;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.ImageWriteParam;
import javax.imageio.ImageWriter;
import javax.imageio.event.IIOReadProgressListener;
import javax.imageio.event.IIOWriteProgressListener;
import javax.imageio.stream.ImageInputStream;
import javax.imageio.stream.ImageOutputStream;

import net.sf.image4j.codec.bmp.BMPEncoder;
import net.sf.image4j.codec.bmp.InfoHeader;
import net.sf.image4j.codec.ico.ICOConstants;
import net.sf.image4j.codec.ico.ICOEncoder;
import net.sf.image4j.codec.ico.IconEntry;
import net.sf.image4j.io.LittleEndianOutputStream;
import net.sf.image4j.util.ConvertUtil;

import org.imgscalr.Scalr;

/**
 * A thread for exporting a project either to a single icon in the .ico format
 * or to multiple image files in the .jpg, .png or .bmp format.  Images are
 * scaled according to the user's selected width/height values and the bit-depth
 * is set accordingly as-well.
 * 
 * 
 * @author Adam T. Ryder http://1337atr.weebly.com
 *
 */
public class ImageExport extends Thread {
	public static final int FORMAT_PNG = 0;
	public static final int FORMAT_JPG = 1;
	public static final int FORMAT_BMP = 2;
	
	private static final int thumbSize = 96;
	
	private static int nextID = 0;
	private static ImageExport singleton;
	
	private boolean cancelled = false;
	private int ID;
	private final boolean isBatch;
	
	private final GuavacadoMainWin mainWin;
	
	private final File path;
	
	private final double quality;
	private final int format;
	
	private IconTile[] tiles;
	private final File[] exportPaths;
	private long inputSizes[];
	private long outputSizes[];
	private long totalSize = 0;
	
	private float percDone = 0f;
	
	/**
	 * Constructs a new instance for use with exporting a project in the
	 * .ico format.
	 * 
	 * @param path <code>File</code> representing the .ico to which to export the
	 * project to.
	 * @param mainWin The main Guavacado window class.
	 */
	private ImageExport(final File path, final GuavacadoMainWin mainWin) {
		isBatch = false;
		quality = 0;
		format = 0;
		ID = nextID;
		nextID = (nextID == Integer.MAX_VALUE) ? 0 : nextID + 1;
		this.mainWin = mainWin;
		this.path = path;
		
		exportPaths = null;
	}
	
	/**
	 * Constructs a new instance for use with exporting the project as multiple
	 * images in a selected format.
	 * 
	 * @param path A <code>File</code> representing the folder in which to export
	 * the images.
	 * @param quality The quality setting for use with jpeg format.
	 * @param format The format to export in such as <code>ImageExport.FORMAT_PNG</code>
	 * @param mainWin The main Guavacado window class.
	 */
	private ImageExport(final File path, final double quality, final int format, final GuavacadoMainWin mainWin) {
		isBatch = true;
		this.quality = quality;
		this.format = format;
		ID = nextID;
		nextID = (nextID == Integer.MAX_VALUE) ? 0 : nextID + 1;
		this.mainWin = mainWin;
		this.path = path;
		
		int count = 0;
		for (IconTile tile : mainWin.getTiles()) {
			if (!tile.isError()) {
				count++;
			}
		}
		
		final String ext;
		switch(format) {
		case FORMAT_JPG:
			ext = ".jpg";
			break;
		case FORMAT_BMP:
			ext = ".bmp";
			break;
		default:
			ext = ".png";
			break;
		}
		tiles = new IconTile[count];
		exportPaths = new File[count];
		inputSizes = new long[count];
		outputSizes = new long[count];
		count = 0;
		final StringBuilder sb = new StringBuilder();
		for (int i = 0; i < mainWin.getTiles().size(); i++) {
			if (!mainWin.getTiles().get(i).isError()) {
				tiles[count] = mainWin.getTiles().get(i);
				
				sb.append(tiles[count].imgSrc.getName().substring(0, tiles[count].imgSrc.getName().lastIndexOf(".")));
				sb.append("_");
				sb.append(Integer.toString(tiles[count].getScaledWidth()));
				if (tiles[count].getScaledWidth() != tiles[count].getScaledHeight()) {
					sb.append("x");
					sb.append(Integer.toString(tiles[count].getScaledHeight()));
				}
				sb.append("_");
				sb.append(Integer.toString((tiles[count].getOutputBpp() > 24 && (format == FORMAT_BMP || format == FORMAT_JPG)) ? 24 : tiles[count].getOutputBpp()));
				sb.append("bit");
				sb.append(ext);
				
				exportPaths[count] = new File(path, sb.toString());
				count++;
				sb.delete(0, sb.length());
				
				inputSizes[i] = ImageLoader.getImageSize(tiles[i].imageWidth, tiles[i].imageHeight, tiles[i].imageBitDepth);
				outputSizes[i] = ImageLoader.getImageSize(tiles[i].getScaledWidth(), tiles[i].getScaledHeight(), tiles[i].getOutputBpp());
				totalSize += inputSizes[i] + outputSizes[i];
			}
		}
		
		boolean noStart = false;
		List<String> names = new ArrayList<String>();
		names.add(exportPaths[0].getName());
		for (int i = 1; i < exportPaths.length; i++) {
			if (!names.contains(exportPaths[i].getName())) {
				names.add(exportPaths[i].getName());
			} else {
				int dupNum = 0;
				sb.append(exportPaths[i].getName().substring(0, exportPaths[i].getName().lastIndexOf(".")));
				int subStart = sb.length();
				while (dupNum < Integer.MAX_VALUE) {
					sb.append("(");
					sb.append(Integer.toString(dupNum));
					sb.append(")");
					sb.append(ext);
					
					if (!names.contains(sb.toString())) {
						exportPaths[i] = new File(path, sb.toString());
						names.add(sb.toString());
						break;
					} else {
						sb.delete(subStart, sb.length());
					}
					dupNum ++;
				}
				sb.delete(0, sb.length());
				if (dupNum == Integer.MAX_VALUE) {
					noStart = true;
					break;
				}
			}
		}
		
		if (noStart) {
			mainWin.popError("Not enough available filenames to complete operation.");
		} else {
			names.clear();
			for (File file : exportPaths) {
				if (file.exists()) {
					names.add(file.getName());
				}
			}
			
			if (names.isEmpty()) {
				start();
			} else {
				mainWin.popOverwrite(ID, names);
			}
		}
	}
	
	/**
	 * Creates a singleton instance of <code>ImageExport</code> for use with
	 * exporting a project in the .ico format and starts the thread.
	 * 
	 * @param saveFile A <code>File</code> representing the icon file in which
	 * to export the project.
	 * @param mainWin The Guavacado main window class.
	 */
	protected static void startIcoExport(final File saveFile, final GuavacadoMainWin mainWin) {
		singleton = new ImageExport(saveFile, mainWin);
		singleton.start();
	}
	
	/**
	 * Creates a singleton instance of <code>ImageExport</code> for use with
	 * exporting a project as multiple files in the .png, .jpg or .bmp format
	 * and starts the thread.
	 * 
	 * @param directory A <code>File</code> representing the folder in which to
	 * save the resulting images.
	 * @param quality The quality setting to be used for jpeg format.
	 * @param format The format to export in such as <code>ImageExport.FORMAT_PNG</code>
	 * @param mainWin The Guavacado main window class.
	 */
	protected static void startBatchExport(final File directory, final double quality, final int format, final GuavacadoMainWin mainWin) {
		singleton = new ImageExport(directory, quality, format, mainWin);
	}
	
	protected static void startBatchExport(final int id) {
		if (singleton != null) {
			if (singleton.ID == id) {
				singleton.start();
			}
		}
	}
	
	protected static void declineBatchExport(final int id) {
		if (singleton != null) {
			if (singleton.ID == id) {
				singleton = null;
				ImageLoader.mainWin.finishProgress();
			}
		}
	}
	
	public boolean isCancelled() {
		return cancelled;
	}
	
	public static void setCancelled() {
		if (singleton != null) {
			singleton.cancel();
		}
	}
	
	private void cancel() {
		cancelled = true;
	}
	
	@Override
	public void run() {
		if (isBatch) {
			batchExport();
		} else {
			icoExport();
		}
		
		Platform.runLater(new Runnable() {
			@Override
			public void run() {
				if (!isCancelled()) {
					singleton = null;
				} else if (singleton.ID == ID) {
					singleton = null;
				}
				mainWin.finishProgress();
			}
		});
	}
	
	/**
	 * Export the project as multiple images.
	 */
	private void batchExport() {
		for (int i = 0; i < tiles.length; i++) {
			final ImageReader reader = ImageIO.getImageReadersBySuffix(ImageLoader.getExtension(tiles[i].imgSrc.getName())).next();
			final ImageWriter writer = ImageIO.getImageWritersBySuffix(ImageLoader.getExtension(exportPaths[i].getName())).next();
			ImageWriteParam iwp = writer.getDefaultWriteParam();
			if (format == FORMAT_JPG) {
				iwp.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
				iwp.setCompressionQuality((float)Math.min(1.0, quality / 100.0));
			}
			
			boolean error = false;
			ImageInputStream ios = null;
			ImageOutputStream iouts= null;
			try {
				ios = ImageIO.createImageInputStream(tiles[i].imgSrc);
				iouts = ImageIO.createImageOutputStream(exportPaths[i]);
				
				reader.setInput(ios);
				reader.addIIOReadProgressListener(new ReadProgress((float)inputSizes[i] / totalSize));
				writer.setOutput(iouts);
				writer.addIIOWriteProgressListener(new WriteProgress((float)outputSizes[i] / totalSize));
				List<BufferedImage> thumb = new ArrayList<BufferedImage>();
				
				if (tiles[i].getScaledWidth() != tiles[i].imageWidth || tiles[i].getScaledHeight() != tiles[i].imageHeight) {
					if ((format == FORMAT_JPG || format == FORMAT_BMP) && tiles[i].getOutputBpp() > 24) {
						if (tiles[i].imageBitDepth != 24) {
							BufferedImage img = convertBPP(Scalr.resize(reader.read(reader.getMinIndex()), Scalr.Method.ULTRA_QUALITY, Scalr.Mode.FIT_EXACT, tiles[i].getScaledWidth(), tiles[i].getScaledHeight()), 24);
							if (img.getWidth() > thumbSize || img.getHeight() > thumbSize) {
								thumb.add(Scalr.resize(img, Scalr.Method.QUALITY, thumbSize));
							} else {
								thumb = null;
							}
							iwp.setDestinationType(new ImageTypeSpecifier(img.getColorModel(), img.getSampleModel()));
							writer.write(null, new IIOImage(img, thumb, null), iwp);
						} else {
							BufferedImage img = Scalr.resize(reader.read(reader.getMinIndex()), Scalr.Method.ULTRA_QUALITY, Scalr.Mode.FIT_EXACT, tiles[i].getScaledWidth(), tiles[i].getScaledHeight());
							if (img.getWidth() > thumbSize || img.getHeight() > thumbSize) {
								thumb.add(Scalr.resize(img, Scalr.Method.QUALITY, thumbSize));
							} else {
								thumb = null;
							}
							iwp.setDestinationType(new ImageTypeSpecifier(img.getColorModel(), img.getSampleModel()));
							writer.write(null, new IIOImage(img, thumb, null), iwp);
						}
					} else {
						if (tiles[i].getOutputBpp() != tiles[i].imageBitDepth) {
							BufferedImage img = convertBPP(Scalr.resize(reader.read(reader.getMinIndex()), Scalr.Method.ULTRA_QUALITY, Scalr.Mode.FIT_EXACT, tiles[i].getScaledWidth(), tiles[i].getScaledHeight()), tiles[i].getOutputBpp());
							if (img.getWidth() > thumbSize || img.getHeight() > thumbSize) {
								thumb.add(Scalr.resize(img, Scalr.Method.QUALITY, thumbSize));
							} else {
								thumb = null;
							}
							if (format == FORMAT_JPG && tiles[i].getOutputBpp() == 4) {
								BufferedImage tmpImage = new BufferedImage(img.getWidth(), img.getHeight(), BufferedImage.TYPE_INT_RGB);
								new ColorConvertOp(img.getColorModel().getColorSpace(), tmpImage.getColorModel().getColorSpace(), null).filter(img, tmpImage);
								img = tmpImage;
							}
							iwp.setDestinationType(new ImageTypeSpecifier(img.getColorModel(), img.getSampleModel()));
							writer.write(null, new IIOImage(img, thumb, null), iwp);
						} else {
							BufferedImage img = Scalr.resize(reader.read(reader.getMinIndex()), Scalr.Method.ULTRA_QUALITY, Scalr.Mode.FIT_EXACT, tiles[i].getScaledWidth(), tiles[i].getScaledHeight());
							if (img.getWidth() > thumbSize || img.getHeight() > thumbSize) {
								thumb.add(Scalr.resize(img, Scalr.Method.QUALITY, thumbSize));
							} else {
								thumb = null;
							}
							if (format == FORMAT_JPG && tiles[i].getOutputBpp() == 4) {
								BufferedImage tmpImage = new BufferedImage(img.getWidth(), img.getHeight(), BufferedImage.TYPE_INT_RGB);
								new ColorConvertOp(img.getColorModel().getColorSpace(), tmpImage.getColorModel().getColorSpace(), null).filter(img, tmpImage);
								img = tmpImage;
							}
							iwp.setDestinationType(new ImageTypeSpecifier(img.getColorModel(), img.getSampleModel()));
							writer.write(null, new IIOImage(img, thumb, null), iwp);
						}
					}
				} else if (tiles[i].getOutputBpp() != tiles[i].imageBitDepth) {
					BufferedImage img = convertBPP(reader.read(reader.getMinIndex()), (tiles[i].getOutputBpp() > 24 && (format == FORMAT_JPG || format == FORMAT_BMP)) ? 24 : tiles[i].getOutputBpp());
					if (img.getWidth() > thumbSize || img.getHeight() > thumbSize) {
						thumb.add(Scalr.resize(img, Scalr.Method.QUALITY, thumbSize));
					} else {
						thumb = null;
					}
					if (format == FORMAT_JPG && tiles[i].getOutputBpp() == 4) {
						BufferedImage tmpImage = new BufferedImage(img.getWidth(), img.getHeight(), BufferedImage.TYPE_INT_RGB);
						new ColorConvertOp(img.getColorModel().getColorSpace(), tmpImage.getColorModel().getColorSpace(), null).filter(img, tmpImage);
						img = tmpImage;
					}
					iwp.setDestinationType(new ImageTypeSpecifier(img.getColorModel(), img.getSampleModel()));
					writer.write(null, new IIOImage(img, thumb, null), iwp);
				} else {
					if ((format == FORMAT_JPG || format == FORMAT_BMP) && tiles[i].getOutputBpp() > 24) {
						BufferedImage img = convertBPP(reader.read(reader.getMinIndex()), 24);
						if (img.getWidth() > thumbSize || img.getHeight() > thumbSize) {
							thumb.add(Scalr.resize(img, Scalr.Method.QUALITY, thumbSize));
						} else {
							thumb = null;
						}
						iwp.setDestinationType(new ImageTypeSpecifier(img.getColorModel(), img.getSampleModel()));
						writer.write(null, new IIOImage(img, thumb, null), iwp);
					} else {
						BufferedImage img = reader.read(reader.getMinIndex());
						if (img.getWidth() > thumbSize || img.getHeight() > thumbSize) {
							thumb.add(Scalr.resize(img, Scalr.Method.QUALITY, thumbSize));
						} else {
							thumb = null;
						}
						if (format == FORMAT_JPG && tiles[i].getOutputBpp() == 4) {
							BufferedImage tmpImage = new BufferedImage(img.getWidth(), img.getHeight(), BufferedImage.TYPE_INT_RGB);
							new ColorConvertOp(img.getColorModel().getColorSpace(), tmpImage.getColorModel().getColorSpace(), null).filter(img, tmpImage);
							img = tmpImage;
						}
						iwp.setDestinationType(new ImageTypeSpecifier(img.getColorModel(), img.getSampleModel()));
						writer.write(null, new IIOImage(img, thumb, null), iwp);
					}
				}
			} catch (Exception e) {
				final String err = e.getMessage();
				error = true;
				Platform.runLater(new Runnable() {
					@Override
					public void run() {
						mainWin.popError("Guavacado encountered the following error:\n" + err);
					}
				});
			} catch (OutOfMemoryError oom) {
				error = true;
				final String err = tiles[i].imgSrc.getName();
				Platform.runLater(new Runnable() {
					@Override
					public void run() {
						mainWin.popError("Guavacado ran out of memory while trying to export " + err + ".");
					}
				});
			} finally {
				reader.dispose();
				writer.dispose();
				try {
					ios.close();
					iouts.close();
				} catch (IOException ioe) {
					
				}
			}
			
			if (cancelled || error) {
				break;
			}
		}
	}
	
	/**
	 * Export the project as a single icon in the .ico format.
	 */
	private void icoExport() {
		int count = 0;
		int compressedCount = 0;
		for (int i = 0; i < mainWin.getTiles().size(); i++) {
			final IconTile tile = mainWin.getTiles().get(i);
			if (!tile.isError() && !cancelled) {
				count++;
				if (tile.pngOutput()) {
					compressedCount++;
				}
			} else if (cancelled) {
				break;
			}
		}
		
		tiles = new IconTile[count];
		inputSizes = new long[count];
		outputSizes = new long[count];
		int num = 0;
		for (int i = 0; i < mainWin.getTiles().size(); i++) {
			final IconTile tile = mainWin.getTiles().get(i);
			if (!tile.isError() && !cancelled) {
				inputSizes[num] = ImageLoader.getImageSize(tile.imageWidth, tile.imageHeight, tile.imageBitDepth);
				outputSizes[num] = ImageLoader.getImageSize(tile.getScaledWidth(), tile.getScaledHeight(), tile.getOutputBppICO());
				totalSize += inputSizes[num] + outputSizes[num];
				tiles[num] = tile;
				num++;
			} else if (cancelled) {
				break;
			}
		}
		
		String error = null;
		if (!cancelled) {
			LittleEndianOutputStream los = null;
			ImageWriter pngWriter = null;
			try {
				los = new LittleEndianOutputStream(new BufferedOutputStream(new FileOutputStream(path)));
				ICOEncoder.writeFileHeader(count, ICOConstants.TYPE_ICON, los);
				int fileOffset = 6 + count * 16;
				
				InfoHeader[] infoHeaders = new InfoHeader[count];
				BufferedImage[] processed = new BufferedImage[count];
				byte[][] andBMP = new byte[count][];
				byte[][] pngBytes = new byte[compressedCount][];
				
				if (compressedCount > 0) {
					pngWriter = ImageIO.getImageWritersBySuffix("png").next();
				}
				
				int compressConverted = 0;
				for (int i = 0; i < tiles.length; i++) {
					if (!cancelled) {
						final IconTile tile = tiles[i];
						error = convertImage(tile, inputSizes[i], totalSize, processed, andBMP, i);
						
						if (error == null) {
							infoHeaders[i] = BMPEncoder.createInfoHeader(processed[i]);
							IconEntry ie = ICOEncoder.createIconEntry(infoHeaders[i]);
							
							if (tile.pngOutput()) {
								ByteArrayOutputStream bout = new ByteArrayOutputStream();
								ImageOutputStream ios = ImageIO.createImageOutputStream(bout);
								pngWriter.setOutput(ios);
								
								pngWriter.write(processed[i]);
								bout.flush();
								pngBytes[compressConverted++] = bout.toByteArray();
								
								ios.close();
								processed[i] = null;
								
								ie.iSizeInBytes = pngBytes[compressConverted - 1].length;
							}
							
							infoHeaders[i].iHeight *= 2;
							ie.iFileOffset = fileOffset;
							fileOffset += ie.iSizeInBytes;
							
							ie.write(los);
						} else {
							final String err = error;
							Platform.runLater(new Runnable() {
								@Override
								public void run() {
									mainWin.popError(err);
								}
							});
							break;
						}
						
					} else { break; }
				}
				
				if (error == null && !cancelled) {
					compressConverted = 0;
					for (int i = 0; i < processed.length; i++) {
						if (cancelled) { break; }
						
						IIOWriteProgressListener prog = new WriteProgress((float)outputSizes[i] / totalSize);
						
						if (!tiles[i].pngOutput()) {
							infoHeaders[i].write(los);
							if (infoHeaders[i].sBitCount <= 8) {
								BMPEncoder.writeColorMap(((IndexColorModel)processed[i].getColorModel()), los);
							}
							
							//write Xor bitmap
							switch(infoHeaders[i].sBitCount) {
							case 1:
								BMPEncoder.write1(processed[i].getRaster(), los);
								break;
							case 4:
								BMPEncoder.write4(processed[i].getRaster(), los);
								break;
							case 8:
								BMPEncoder.write8(processed[i].getRaster(), los);
								break;
							case 24:
								BMPEncoder.write24(processed[i].getRaster(), los);
								break;
							case 32:
								BMPEncoder.write32(processed[i].getRaster(), processed[i].getAlphaRaster(), los);
								break;
							}
							
							prog.imageProgress(null, 0.5f);
							
							//write And bitmap
							if (andBMP[i] != null) {
								los.write(andBMP[i]);
							} else {
								byte[] line = new byte[getBytesPerLine1(processed[i].getWidth())];
								
								for (int byi = 0; byi < line.length; byi++) {
									line[byi] = (byte)0;
								}
								
								for (int y = processed[i].getHeight() - 1; y >= 0; y--) {
									los.write(line);
								}
							}
						} else {
							los.write(pngBytes[compressConverted++]);
						}
						
						prog.imageComplete(null);
					}
				}
				
			} catch (OutOfMemoryError oom) {
				Platform.runLater(new Runnable() {
					@Override
					public void run() {
						mainWin.popError("There is not enough available system memory to complete this process.");
					}
				});
			} catch (IOException ioE) {
				final String err = ioE.getMessage();
				Platform.runLater(new Runnable() {
					@Override
					public void run() {
						mainWin.popError("Guavacado encountered the following error:\n" + err);
					}
				});
			} finally {
				try {
					los.close();
					if (pngWriter != null) {
						pngWriter.dispose();
					}
				} catch (IOException ioe) {
					
				}
			}
		}
	}
	
	/**
	 * Resize and convert bit-depth of a project image.
	 * 
	 * @param tile The <code>ImageTile</code> containing the image and settings to
	 * use for conversion.
	 * @param inputSize Size of the input image in Bytes.
	 * @param totalSize total size of all input and output images in Bytes.
	 * @param images Array of <code>BufferedImages</code> in which to store
	 * the converted images.
	 * @param andBMP Array of <code>byte[]</code> to store all the icon AND maps.
	 * @param index Index of the image within the <code>BufferedImages</code> array.
	 * @return A string containing any errors generated during the process or
	 * null if no errors were encountered.
	 */
	private String convertImage(final IconTile tile, final long inputSize, final long totalSize, final BufferedImage[] images, final byte[][] andBMP, final int index) {
		String error = null;
		
		ImageReader reader = ImageIO.getImageReadersBySuffix(ImageLoader.getExtension(tile.imgSrc.getName())).next();
		reader.addIIOReadProgressListener(new ReadProgress((float)inputSize / totalSize));
		ImageInputStream ios = null;
		try {
			ios = ImageIO.createImageInputStream(tile.imgSrc);
			reader.setInput(ios);
			if (tile.getScaledHeight() != tile.imageHeight || tile.getScaledWidth() != tile.imageWidth) {
				if (tile.getOutputBppICO() != tile.imageBitDepth) {
					BufferedImage original = Scalr.resize(reader.read(reader.getMinIndex()), Scalr.Method.ULTRA_QUALITY, Scalr.Mode.FIT_EXACT, tile.getScaledWidth(), tile.getScaledHeight());
					andBMP[index] = getANDMap(original);
					images[index] = convertBPP(original, tile.getOutputBppICO());
				} else {
					images[index] = Scalr.resize(reader.read(reader.getMinIndex()), Scalr.Method.ULTRA_QUALITY, Scalr.Mode.FIT_EXACT, tile.getScaledWidth(), tile.getScaledHeight());
					andBMP[index] = getANDMap(images[index]);
				}
			} else {
				if (tile.getOutputBppICO() != tile.imageBitDepth) {
					BufferedImage original = reader.read(reader.getMinIndex());
					andBMP[index] = getANDMap(original);
					images[index] = convertBPP(original, tile.getOutputBppICO());
				} else {
					images[index] = reader.read(reader.getMinIndex());
					andBMP[index] = getANDMap(images[index]);
				}
			}
		} catch (IOException convertIOE) {
			error = convertIOE.getMessage();
			System.out.println(error);
		} finally {
			if (ios != null) {
				try {
					ios.close();
				} catch (IOException iose) {
					
				}
			}
			reader.dispose();
		}
		
		return error;
	}
	
	/**
	 * Creates an AND map for use with the .ico format.  Much of this code is merely
	 * modified from code provided by the Image4j library written by Ian McDonagh.
	 * 
	 * @param bimg The <code>Buffered</code> image for which the AND map should
	 * be generated.
	 * @return a <code>byte[]</code> containing the AND map bytes.
	 */
	private byte[] getANDMap(final BufferedImage bimg) {
		WritableRaster alpha = bimg.getAlphaRaster();
		if (bimg.getColorModel() instanceof IndexColorModel && bimg.getColorModel().hasAlpha()) {
			int bytesPerLine = getBytesPerLine1(bimg.getWidth());
			byte[] line = new byte[bytesPerLine * bimg.getHeight()];
			
			IndexColorModel icm = (IndexColorModel)bimg.getColorModel();
			Raster raster = bimg.getRaster();
			
			for (int y = bimg.getHeight() - 1; y >= 0; y--) {
				for (int x = 0; x < bimg.getWidth(); x++) {
					int bi = (x / 8) + (((bimg.getHeight() - 1) - y) * bytesPerLine);
					int i = x % 8;
					int p = raster.getSample(x, y, 0);
					int a = icm.getAlpha(p);
					int b = ~a & 1;
					
					line[bi] &= ~(1 << (7 - i));
					line[bi] |= b << (7 - i);
				}
			}
			
			return line;
		} else if (alpha != null) {
			int bytesPerLine = getBytesPerLine1(bimg.getWidth());
			byte[] line = new byte[bytesPerLine * bimg.getHeight()];
			
			for (int y = bimg.getHeight() - 1; y >= 0; y--) {
				for (int x = 0; x < bimg.getWidth(); x++) {
					int bi = (x / 8) + (((bimg.getHeight() - 1) - y) * bytesPerLine);
					int i = x % 8;
					int a = alpha.getSample(x, y, 0);
					int b = ~a & 1;
					
					line[bi] &= ~(1 << (7 - i));
					line[bi] |= b << (7 - i);
				}
			}
			
			return line;
		}
		
		return null;
	}
	
	/**
	 * Slightly modified code originally written by Ian McDonagh for the
	 * Image4j library.
	 */
	private int getBytesPerLine1(int width) {
		int ret = (int) Math.ceil((float)width / 8);
	    if (ret % 4 != 0) {
	      ret = (ret / 4 + 1) * 4;
	    }

	    return ret;
	}
	
	/**
	 * Mostly uses the Image4j library written by Ian McDonagh, but adds in
	 * a 16-bit conversion.
	 * 
	 * @param image The <code>BufferedImage</code> to convert.
	 * @param bpp The desired bit depth.
	 * @return The converted <code>BufferedImage</code>.
	 */
	private BufferedImage convertBPP(final BufferedImage image, final int bpp) {
		switch(bpp) {
		case 1:
			return ConvertUtil.convert1(image);
		case 4:
			return ConvertUtil.convert4(image);
		case 8:
			return ConvertUtil.convert8(image);
		case 16:
			BufferedImage dest = new BufferedImage(image.getWidth(), image.getHeight(), BufferedImage.TYPE_USHORT_565_RGB);
			Graphics2D g2d = (Graphics2D)dest.getGraphics();
			g2d.setComposite(AlphaComposite.Src);
			g2d.drawImage(image, 0, 0, null);
			return dest;
		case 24:
			return ConvertUtil.convert24(image);
		default: //32
			return ConvertUtil.convert32(image);
		}
	}
	
	/**
	 * <code>IIOReadProgressListener</code> used for displaying the progress of image
	 * read operations when reading an image from disk.
	 * 
	 * @author Adam T. Ryder http://1337atr.weebly.com
	 *
	 */
	private class ReadProgress implements IIOReadProgressListener {
		private final float percModifier;
		private final StringBuilder sb = new StringBuilder("0%");
		private long lastCheck;
		
		private ReadProgress(final double percModifier) {
			this.percModifier = (float)percModifier;
			lastCheck = System.nanoTime();
		}
		
		@Override
		public void imageComplete(ImageReader source) {
			Platform.runLater(new Runnable() {
				@Override
				public void run() {
					if (!isCancelled()) {
						percDone += 100f * percModifier;
						sb.delete(0, sb.length());
						sb.append(Integer.toString((int)Math.floor(percDone)));
						sb.append("%");
						
						mainWin.updateProgress(sb.toString(), percDone / 100);
					}
				}
			});
		}
		
		@Override
		public void imageProgress(ImageReader source, float percentDone) {
			if (System.nanoTime() - lastCheck >= 1000000000 * 0.04167) {
				lastCheck = System.nanoTime();
				Platform.runLater(new Runnable() {
					@Override
					public void run() {
						if (!isCancelled()) {
							float perc = Math.min(100, percDone + (percentDone * percModifier));
							sb.delete(0, sb.length());
							sb.append(Integer.toString((int)Math.round(perc)));
							sb.append("%");
							mainWin.updateProgress(sb.toString(), perc / 100);
						}
					}
				});
			}
		}
		
		@Override
		public void imageStarted(ImageReader source, int imageIndex) {
			
		}
		
		@Override
		public void readAborted(ImageReader source) {
			
		}
		
		@Override
		public void sequenceComplete(ImageReader source) {
			
		}
		
		@Override
		public void sequenceStarted(ImageReader source, int minIndex) {
			
		}
		
		@Override
		public void thumbnailComplete(ImageReader source) {
			
		}
		
		@Override
		public void thumbnailProgress(ImageReader source, float percentDone) {
			
		}
		
		@Override
		public void thumbnailStarted(ImageReader source, int imageIndex, int thumbnailIndex) {
			
		}
	}
	
	/**
	 * <code>IIOWriteProgressListener</code> used for displaying the progress of image
	 * write operations when writing an image to disk.
	 * 
	 * @author Adam T. Ryder http://1337atr.weebly.com
	 *
	 */
	private class WriteProgress implements IIOWriteProgressListener {
		private final float percModifier;
		private final StringBuilder sb = new StringBuilder("0%");
		private long lastCheck;
		
		private WriteProgress(final double percModifier) {
			this.percModifier = (float)percModifier;
			lastCheck = System.nanoTime();
		}
		
		@Override
		public void imageComplete(ImageWriter source) {
			Platform.runLater(new Runnable() {
				@Override
				public void run() {
					if (!isCancelled()) {
						percDone += 100f * percModifier;
						sb.delete(0, sb.length());
						sb.append(Integer.toString((int)Math.floor(percDone)));
						sb.append("%");
						
						mainWin.updateProgress(sb.toString(), percDone / 100);
					}
				}
			});
		}
		
		@Override
		public void imageProgress(ImageWriter source, float percentDone) {
			if (System.nanoTime() - lastCheck >= 1000000000 * 0.04167) {
				lastCheck = System.nanoTime();
				Platform.runLater(new Runnable() {
					@Override
					public void run() {
						if (!isCancelled()) {
							float perc = Math.min(100, percDone + (percentDone * percModifier));
							sb.delete(0, sb.length());
							sb.append(Integer.toString((int)Math.round(perc)));
							sb.append("%");
							mainWin.updateProgress(sb.toString(), perc / 100);
						}
					}
				});
			}
		}
		
		@Override
		public void imageStarted(ImageWriter source, int imageIndex) {
			
		}
		
		@Override
		public void writeAborted(ImageWriter source) {
			
		}
		
		@Override
		public void thumbnailComplete(ImageWriter source) {
			
		}
		
		@Override
		public void thumbnailProgress(ImageWriter source, float percentDone) {
			
		}
		
		@Override
		public void thumbnailStarted(ImageWriter source, int imageIndex, int thumbnailIndex) {
			
		}
	}
}
