package guavacado;

import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javafx.application.Platform;

import javax.imageio.ImageIO;
import javax.imageio.ImageReader;
import javax.imageio.event.IIOReadProgressListener;
import javax.imageio.stream.ImageInputStream;

import org.imgscalr.Scalr;

/**
 * A thread for loading and scaling images for display as a thumbnail on an
 * <code>ImageTile</code>.
 * 
 * @author Adam T. Ryder http://1337atr.weebly.com
 *
 */
public class ImageLoader extends Thread {
	private static ImageLoader singleton;
	private static final List<IconTile> tiles = new ArrayList<IconTile>();
	private static final List<IconTile> pendingTiles = new ArrayList<IconTile>();
	
	private boolean cancelled = false;
	
	public static GuavacadoMainWin mainWin;
	
	private ImageLoader() {
		
	}
	
	/**
	 * Add an <code>ImageTile</code> to be loaded.
	 * 
	 * @param tile The <code>ImageTile</code> that needs a new thumbnail and
	 * image data.
	 */
	protected static void addImage(final IconTile tile) {
		if (singleton == null) {
			tiles.add(tile);
			singleton = new ImageLoader();
			singleton.start();
		} else {
			updateTiles(tile);
		}
	}
	
	/**
	 * Add an <code>ImageTile</code> to be loaded.
	 * 
	 * @param tile The <code>ImageTile</code> that needs a new thumbnail and
	 * image data.
	 */
	private static synchronized void updateTiles(final IconTile tile) {
		if (tile != null) {
			pendingTiles.add(tile);
		} else if (!pendingTiles.isEmpty()) {
			tiles.addAll(pendingTiles);
			pendingTiles.clear();
		}
	}
	
	public boolean isCancelled() {
		return cancelled;
	}
	
	public static void setCancelled() {
		if (singleton != null) {
			singleton.cancel();
		}
	}
	
	private void cancel() {
		cancelled = true;
	}
	
	/**
	 * Gets the extension from a supplied filename/path for use in determining
	 * the image format.
	 * 
	 * @param filePath
	 * @return Lowercase image format, jpg is returned if the extension is jpeg.
	 */
	public static String getExtension(final String filePath) {
		if (filePath.toLowerCase().endsWith(".jpeg")) {
			return "jpg";
		}
		int index = filePath.lastIndexOf(".");
		
		return filePath.substring(index + 1).toLowerCase();
	}
	
	/**
	 * Returns the size of an image in bytes.
	 * 
	 * @param width Width of the image.
	 * @param height Height of the image.
	 * @param bitDepth bit depth of the image.
	 * @return Size of the image in bytes.
	 */
	public static long getImageSize(int width, int height, int bitDepth) {
		return ((width * height) * bitDepth) / 8;
	}
	
	@Override
	public void run() {
		while (!tiles.isEmpty()) {
			
			Iterator<IconTile> iterator = tiles.iterator();
			while (iterator.hasNext() && !cancelled) {
				final IconTile tile = iterator.next();
				
				if (!tile.isCancelled()) {
					ImageInputStream ios = null;
					ImageReader reader = ImageIO.getImageReadersBySuffix(getExtension(tile.imgSrc.getName())).next();
					reader.addIIOReadProgressListener(new LoadProgress(tile));
					final BufferedImage bi;
					try {
						ios = ImageIO.createImageInputStream(tile.imgSrc);
						reader.setInput(ios);
						final int minImgIndex = reader.getMinIndex();
						final int imgWidth = reader.getWidth(minImgIndex);
						final int imgHeight = reader.getHeight(minImgIndex);
						final int imgBitDepth = reader.getImageTypes(minImgIndex).next().getColorModel().getPixelSize();
						
						try {
							if (imgWidth > IconTile.displayWidth || imgHeight > IconTile.displayHeight) {
								bi = Scalr.resize(reader.read(minImgIndex), Scalr.Method.QUALITY, IconTile.displayWidth);
							} else {
								bi = reader.read(minImgIndex);
							}
							Platform.runLater(new Runnable() {
								@Override
								public void run() {
									if (!tile.isCancelled()) {
										tile.setImage(bi, imgWidth, imgHeight, imgBitDepth);
									}
								}
							});
						} catch (OutOfMemoryError oom) {
							Platform.runLater(new Runnable() {
								@Override
								public void run() {
									if (!tile.isCancelled()) {
										tile.setOutOfMemory(imgWidth, imgHeight, imgBitDepth);
									}
								}
							});
						}
					} catch (Exception e) {
						Platform.runLater(new Runnable() {
							@Override
							public void run() {
								if (!tile.isCancelled()) {
									tile.setError();
								}
							}
						});
					} finally {
						reader.dispose();
						try {
							ios.close();
						} catch (IOException ioe) {
							
						}
					}
				}
				
				iterator.remove();
			}
			
			updateTiles(null);
		}
		
		Platform.runLater(new Runnable() {
			@Override
			public void run() {
				if (pendingTiles.isEmpty()) {
					singleton = null;
				} else {
					tiles.addAll(pendingTiles);
					pendingTiles.clear();
					singleton = new ImageLoader();
					singleton.start();
				}
			}
		});
	}
	
	private class LoadProgress implements IIOReadProgressListener {
		private final IconTile tile;
		private final StringBuilder sb = new StringBuilder("0%");
		long lastCheck;
		
		private LoadProgress(final IconTile tile) {
			this.tile = tile;
			lastCheck = System.nanoTime();
		}
		
		@Override
		public void imageComplete(ImageReader source) {
			Platform.runLater(new Runnable() {
				@Override
				public void run() {
					if (!tile.isCancelled() && !tile.isLoaded() && !isCancelled()) {
						tile.progressLabel.setText("100%");
						tile.arcProgress.setLength(-360);
					}
				}
			});
		}
		
		@Override
		public void imageProgress(ImageReader source, float percentDone) {
			if (System.nanoTime() - lastCheck >= 1000000000 * 0.04167) {
				lastCheck = System.nanoTime();
				Platform.runLater(new Runnable() {
					@Override
					public void run() {
						if (!tile.isCancelled() && !tile.isLoaded() && !isCancelled()) {
							sb.delete(0, sb.length());
							sb.append(Integer.toString((int)Math.floor(percentDone)));
							sb.append("%");
							tile.progressLabel.setText(sb.toString());
							
							tile.arcProgress.setLength(-360 * (percentDone / 100));
						}
					}
				});
			}
		}
		
		@Override
		public void imageStarted(ImageReader source, int imageIndex) {
			
		}
		
		@Override
		public void readAborted(ImageReader source) {
			
		}
		
		@Override
		public void sequenceComplete(ImageReader source) {
			
		}
		
		@Override
		public void sequenceStarted(ImageReader source, int minIndex) {
			
		}
		
		@Override
		public void thumbnailComplete(ImageReader source) {
			Platform.runLater(new Runnable() {
				@Override
				public void run() {
					if (!tile.isCancelled() && !tile.isLoaded() && !isCancelled()) {
						tile.progressLabel.setText("100%");
						tile.arcProgress.setLength(-360);
					}
				}
			});
		}
		
		@Override
		public void thumbnailProgress(ImageReader source, float percentDone) {
			if (System.nanoTime() - lastCheck >= 1000000000 * 0.04167) {
				lastCheck = System.nanoTime();
				Platform.runLater(new Runnable() {
					@Override
					public void run() {
						if (!tile.isCancelled() && !tile.isLoaded() && !isCancelled()) {
							sb.delete(0, sb.length());
							sb.append(Integer.toString((int)Math.floor(percentDone)));
							sb.append("%");
							tile.progressLabel.setText(sb.toString());
							
							tile.arcProgress.setLength(-360 * (percentDone / 100));
						}
					}
				});
			}
		}
		
		@Override
		public void thumbnailStarted(ImageReader source, int imageIndex, int thumbnailIndex) {
			
		}
	}
}
