package guavacado;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;

import javafx.application.Platform;

/**
 * A <code>Thread</code> used for loading a saved Guavacado Project.
 * 
 * @author Adam T. Ryder http://1337atr.weebly.com
 *
 */
public class ProjectOpen extends Thread {
	private static int nextID = 0;
	
	public static ProjectOpen singleton;
	
	private final GuavacadoMainWin mainWin;
	private final File openPath;
	
	private boolean cancelled = false;
	
	private final int ID;
	
	private projectImage[] entries = null;
	
	private boolean error = false;
	
	private ProjectOpen(final GuavacadoMainWin mainWin, final File openProject) {
		this.mainWin = mainWin;
		openPath = openProject;
		
		ID = nextID;
		nextID = (nextID == Integer.MAX_VALUE) ? 0 : nextID + 1;
	}
	
	/**
	 * Creates a singleton instance of <code>ProjectOpen</code>.
	 * 
	 * @param mainWin The main Guavacado window class instance.
	 * @param projectFile A <code>File</code> representing the saved Guavacado
	 * project file.
	 */
	public static void openProject(final GuavacadoMainWin mainWin, File projectFile) {
		singleton = new ProjectOpen(mainWin, projectFile);
		singleton.start();
	}
	
	public static void setCancelled() {
		if (singleton != null) {
			singleton.cancel();
		}
	}
	
	private void cancel() {
		cancelled = true;
	}
	
	public boolean isCancelled() {
		return cancelled;
	}
	
	@Override
	public void run() {
		long lastCheck = System.nanoTime();
		final StringBuilder sb = new StringBuilder("0%");
		
		DataInputStream dis = null;
		try {
			dis = new DataInputStream(new BufferedInputStream(new FileInputStream(openPath)));
			
			float version = dis.readFloat();
			int numEntries = dis.readInt();
			entries = new projectImage[numEntries];
			for (int i = 0; i < numEntries; i++) {
				entries[i] = new projectImage(dis);
				
				if (System.nanoTime() - lastCheck >= (1000000000 * 0.04167) || i == numEntries - 1) {
					lastCheck = System.nanoTime();
					final float perc = (float)(i + 1) / entries.length;
					sb.delete(0, sb.length());
					sb.append(Integer.toString((int)Math.floor(perc * 100)));
					sb.append("%");
					Platform.runLater(new Runnable() {
						@Override
						public void run() {
							mainWin.updateProgress(sb.toString(), perc);
						}
					});
				}
			}
		} catch (Exception e) {
			error = true;
			Platform.runLater(new Runnable() {
				@Override
				public void run() {
					mainWin.popError("Guavacado was unable to open " + openPath.getName() + " due to the following error:\n" + e.getMessage());
				}
			});
		} finally {
			if (dis != null) {
				try {
					dis.close();
				} catch (IOException ioe) {
					
				}
			}
		}
		
		Platform.runLater(new Runnable() {
			@Override
			public void run() {
				if (singleton.ID == ID) {
					singleton = null;
				}
				if (!isCancelled()) {
					if (!error) {
						mainWin.finishProgress(entries, openPath);
					} else {
						mainWin.finishProgress();
					}
				}
			}
		});
	}
	
	/**
	 * Stores information regarding the settings for an image in a saved
	 * Guavacado project file.
	 * 
	 * @author Adam T. Ryder http://1337atr.weebly.com
	 *
	 */
	public class projectImage {
		public final File path;
		public final int width;
		public final int height;
		public final int bpp;
		public final boolean compress;
		public final boolean isResLocked;
		
		public final boolean isGood;
		
		private projectImage(final DataInputStream dis) throws IOException {
			path = new File(dis.readUTF());
			width = dis.readInt();
			height = dis.readInt();
			bpp = dis.readInt();
			compress = dis.readBoolean();
			isResLocked = dis.readBoolean();
			
			isGood = ((bpp == 1 || bpp == 4 || bpp == 8 || bpp == 16 || bpp == 24 || bpp == 32) && path.exists());
		}
	}
}
