package guavacado;

import java.io.BufferedOutputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

import javafx.application.Platform;

/**
 * A <code>Thread</code> for saving a Guavacado project to disk.
 * 
 * @author Adam T. Ryder http://1337atr.weebly.com
 *
 */
public class ProjectSave extends Thread {
	private static int nextID = 0;
	
	public static ProjectSave singleton;
	
	private final GuavacadoMainWin mainWin;
	private final IconTile[] entries;
	private final File savePath;
	
	private boolean cancelled = false;
	
	private final int ID;
	
	private ProjectSave(final GuavacadoMainWin mainWin) {
		this.mainWin = mainWin;
		savePath = mainWin.getCurrentProjectFile();
		entries = new IconTile[mainWin.getTiles().size()];
		for (int i = 0; i < entries.length; i++) {
			entries[i] = mainWin.getTiles().get(i);
		}
		
		ID = nextID;
		nextID = (nextID == Integer.MAX_VALUE) ? 0 : nextID + 1;
	}
	
	/**
	 * Creates a singleton instance of <code>ProjectSave</code>.
	 * 
	 * @param mainWin The active instance of the Guavacado main window.
	 */
	public static void saveProject(final GuavacadoMainWin mainWin) {
		singleton = new ProjectSave(mainWin);
		singleton.start();
	}
	
	public static void setCancelled() {
		if (singleton != null) {
			singleton.cancel();
		}
	}
	
	private void cancel() {
		cancelled = true;
	}
	
	public boolean isCancelled() {
		return cancelled;
	}
	
	@Override
	public void run() {
		long lastCheck = System.nanoTime();
		final StringBuilder sb = new StringBuilder("0%");
		
		DataOutputStream dos = null;
		try {
			dos = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(savePath)));
			
			dos.writeFloat(mainWin.VERSION);
			dos.writeInt(entries.length);
			for (int i = 0; i < entries.length; i++) {
				if (cancelled) { break; }
				if (!entries[i].isError() && !entries[i].isCancelled()) {
					dos.writeUTF(entries[i].imgSrc.getPath());
					dos.writeInt((entries[i].getScaledWidth() == entries[i].imageWidth) ? -1 : entries[i].getScaledWidth());
					dos.writeInt((entries[i].getScaledHeight() == entries[i].imageHeight) ? -1 : entries[i].getScaledHeight());
					dos.writeInt(entries[i].getOutputBpp());
					dos.writeBoolean(entries[i].pngOutput());
					dos.writeBoolean(entries[i].lockRes.isSelected());
				}
				
				if (System.nanoTime() - lastCheck >= (1000000000 * 0.04167) || i == entries.length - 1) {
					lastCheck = System.nanoTime();
					final float perc = (float)(i + 1) / entries.length;
					sb.delete(0, sb.length());
					sb.append(Integer.toString((int)Math.floor(perc * 100)));
					sb.append("%");
					Platform.runLater(new Runnable() {
						@Override
						public void run() {
							mainWin.updateProgress(sb.toString(), perc);
						}
					});
				}
			}
		} catch (IOException e) {
			Platform.runLater(new Runnable() {
				@Override
				public void run() {
					mainWin.popError("Guavacado was unable to save the project due to the following error:\n" + e.getMessage());
				}
			});
		} finally {
			if (dos != null) {
				try {
					dos.close();
				} catch (IOException ioe) {
					
				}
			}
		}
		
		Platform.runLater(new Runnable() {
			@Override
			public void run() {
				if (singleton.ID == ID) {
					singleton = null;
				}
				if (!isCancelled()) {
					mainWin.finishProgress();
				}
			}
		});
	}
}
